"""Implements the :stac-ext:`Table Extension <table>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar, cast

import pystac
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.utils import get_required

#: Generalized version of :class:`~pystac.Collection`, :class:`~pystac.Item`,
#: :class:`~pystac.Asset` or :class:`~pystac.ItemAssetDefinition`
T = TypeVar(
    "T", pystac.Collection, pystac.Item, pystac.Asset, pystac.ItemAssetDefinition
)

SCHEMA_URI = "https://stac-extensions.github.io/table/v1.2.0/schema.json"

PREFIX: str = "table:"
COLUMNS_PROP = PREFIX + "columns"
PRIMARY_GEOMETRY_PROP = PREFIX + "primary_geometry"
ROW_COUNT_PROP = PREFIX + "row_count"
STORAGE_OPTIONS_PROP = PREFIX + "storage_options"
TABLES_PROP = PREFIX + "tables"

# Column properties
COL_NAME_PROP = "name"
COL_DESCRIPTION_PROP = "description"
COL_TYPE_PROP = "type"

# Table properties
TBL_NAME_PROP = "name"
TBL_DESCRIPTION_PROP = "description"


class Column:
    """Object representing a column of a table."""

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    @property
    def name(self) -> str:
        """The column name"""
        return cast(
            str,
            get_required(
                self.properties.get(COL_NAME_PROP), "table:column", COL_NAME_PROP
            ),
        )

    @name.setter
    def name(self, v: str) -> None:
        self.properties[COL_NAME_PROP] = v

    @property
    def description(self) -> str | None:
        """Detailed multi-line description to explain the column. `CommonMark 0.29
        <http://commonmark.org/>`__ syntax MAY be used for rich text representation."""
        return self.properties.get(COL_DESCRIPTION_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(COL_DESCRIPTION_PROP, None)
        else:
            self.properties[COL_DESCRIPTION_PROP] = v

    @property
    def col_type(self) -> str | None:
        """Data type of the column. If using a file format with a type system (like
        Parquet), we recommend you use those types"""
        return self.properties.get(COL_TYPE_PROP)

    @col_type.setter
    def col_type(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(COL_TYPE_PROP, None)
        else:
            self.properties[COL_TYPE_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representing this ``Column``."""
        return self.properties


class Table:
    """Object containing a high-level summary about a table"""

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    @property
    def name(self) -> str:
        """The table name"""
        return cast(
            str, get_required(self.properties.get(TBL_NAME_PROP), self, TBL_NAME_PROP)
        )

    @name.setter
    def name(self, v: str) -> None:
        self.properties[COL_NAME_PROP] = v

    @property
    def description(self) -> str | None:
        """Detailed multi-line description to explain the table. `CommonMark 0.29
        <http://commonmark.org/>`__ syntax MAY be used for rich text representation."""
        return self.properties.get(COL_DESCRIPTION_PROP)

    @description.setter
    def description(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(COL_DESCRIPTION_PROP, None)
        else:
            self.properties[COL_DESCRIPTION_PROP] = v

    def to_dict(self) -> dict[str, Any]:
        """Returns a dictionary representing this ``Table``."""
        return self.properties


class TableExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of a
    :class:`~pystac.Collection`, :class:`~pystac.Item`, or :class:`~pystac.Asset` with
    properties from the :stac-ext:`Datacube Extension <datacube>`. This class is
    generic over the type of STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Asset`).

    To create a concrete instance of :class:`TableExtension`, use the
    :meth:`TableExtension.ext` method. For example:

    .. code-block:: python

        >>> item: pystac.Item = ...
        >>> tbl_ext = TableExtension.ext(item)

    """

    name: Literal["table"] = "table"

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> TableExtension[T]:
        """Extend the given STAC Object with properties from the
        :stac-ext:`Table Extension <table>`.

        This extension can be applied to instances of :class:`~pystac.Collection`,
        :class:`~pystac.Item` or :class:`~pystac.Asset`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Collection):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(TableExtension[T], CollectionTableExtension(obj))
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(TableExtension[T], ItemTableExtension(obj))
        if isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(TableExtension[T], AssetTableExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(TableExtension[T], ItemAssetsTableExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @property
    def columns(self) -> list[Column] | None:
        """A list of :class:`Column` objects describing each column"""
        v = self.properties.get(COLUMNS_PROP)
        if v is None:
            return None
        return [Column(x) for x in v]

    @columns.setter
    def columns(self, v: list[Column] | None) -> None:
        self._set_property(COLUMNS_PROP, v)

    @property
    def primary_geometry(self) -> str | None:
        """The primary geometry column name"""
        return self._get_property(PRIMARY_GEOMETRY_PROP, str)

    @primary_geometry.setter
    def primary_geometry(self, v: str | None) -> None:
        if v is None:
            self.properties.pop(PRIMARY_GEOMETRY_PROP, None)
        else:
            self.properties[PRIMARY_GEOMETRY_PROP] = v

    @property
    def row_count(self) -> int | None:
        """The number of rows in the dataset"""
        return self._get_property(ROW_COUNT_PROP, int)

    @row_count.setter
    def row_count(self, v: int | None) -> None:
        if v is None:
            self.properties.pop(ROW_COUNT_PROP, None)
        else:
            self.properties[ROW_COUNT_PROP] = v


class CollectionTableExtension(TableExtension[pystac.Collection]):
    """A concrete implementation of :class:`TableExtension` on a
    :class:`~pystac.Collection` that extends the properties of the Item to include
    properties defined in the :stac-ext:`Table Extension <table>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`TableExtension.ext` on an :class:`~pystac.Collection` to extend it.
    """

    collection: pystac.Collection
    properties: dict[str, Any]

    def __init__(self, collection: pystac.Collection):
        self.collection = collection
        self.properties = collection.extra_fields

    @property
    def tables(self) -> dict[str, Table]:
        """A mapping of table names to table objects"""
        return get_required(self.properties.get(TABLES_PROP), self, TABLES_PROP)

    @tables.setter
    def tables(self, v: dict[str, Table]) -> None:
        self.properties[TABLES_PROP] = v

    def __repr__(self) -> str:
        return f"<CollectionTableExtension Item id={self.collection.id}>"


class ItemTableExtension(TableExtension[pystac.Item]):
    """A concrete implementation of :class:`TableExtension` on an
    :class:`~pystac.Item` that extends the properties of the Item to include properties
    defined in the :stac-ext:`Table Extension <table>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`TableExtension.ext` on an :class:`~pystac.Item` to extend it.
    """

    item: pystac.Item
    properties: dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemTableExtension Item id={self.item.id}>"


class AssetTableExtension(TableExtension[pystac.Asset]):
    """A concrete implementation of :class:`TableExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`Table Extension <table>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`TableExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    properties: dict[str, Any]
    additional_read_properties: Iterable[dict[str, Any]] | None

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]
        else:
            self.additional_read_properties = None

    @property
    def storage_options(self) -> dict[str, Any] | None:
        """Additional keywords for opening the dataset"""
        return self.properties.get(STORAGE_OPTIONS_PROP)

    @storage_options.setter
    def storage_options(self, v: dict[str, Any] | None) -> Any:
        if v is None:
            self.properties.pop(STORAGE_OPTIONS_PROP, None)
        else:
            self.properties[STORAGE_OPTIONS_PROP] = v

    def __repr__(self) -> str:
        return f"<AssetTableExtension Item id={self.asset_href}>"


class ItemAssetsTableExtension(TableExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class TableExtensinoHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "table",
        "https://stac-extensions.github.io/table/v1.0.0/schema.json",
        "https://stac-extensions.github.io/table/v1.0.1/schema.json",
        "https://stac-extensions.github.io/table/v1.1.0/schema.json",
    }
    stac_object_types = {pystac.STACObjectType.COLLECTION, pystac.STACObjectType.ITEM}


TABLE_EXTENSION_HOOKS: ExtensionHooks = TableExtensinoHooks()



"""Tests for pystac.tests.extensions.mgrs"""

import json

import pytest

import pystac
from pystac.extensions.mgrs import MgrsExtension
from tests.conftest import get_data_file


@pytest.fixture
def ext_item_uri() -> str:
    return get_data_file("mgrs/item.json")


@pytest.fixture
def ext_item(ext_item_uri: str) -> pystac.Item:
    return pystac.Item.from_file(ext_item_uri)


def test_stac_extensions(ext_item: pystac.Item) -> None:
    assert MgrsExtension.has_extension(ext_item)


def test_get_schema_uri(ext_item: pystac.Item) -> None:
    assert MgrsExtension.get_schema_uri() in ext_item.stac_extensions


def test_ext_raises_if_item_does_not_conform(item: pystac.Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        MgrsExtension.ext(item)


def test_ext_raises_on_collection(collection: pystac.Collection) -> None:
    with pytest.raises(
        pystac.errors.ExtensionTypeError,
        match="MgrsExtension does not apply to type 'Collection'",
    ) as e:
        MgrsExtension.ext(collection)  # type: ignore
    assert "Hint" not in str(e.value)


def test_to_from_dict(ext_item_uri: str, ext_item: pystac.Item) -> None:
    with open(ext_item_uri) as f:
        d = json.load(f)
    actual = ext_item.to_dict(include_self_link=False)
    assert actual == d


def test_add_to(item: pystac.Item) -> None:
    assert not MgrsExtension.has_extension(item)
    MgrsExtension.add_to(item)

    assert MgrsExtension.has_extension(item)


def test_apply(item: pystac.Item) -> None:
    MgrsExtension.add_to(item)
    MgrsExtension.ext(item).apply(latitude_band="X", grid_square="DH")
    assert MgrsExtension.ext(item).latitude_band
    assert MgrsExtension.ext(item).grid_square


def test_apply_without_required_fields_raises(item: pystac.Item) -> None:
    MgrsExtension.add_to(item)
    with pytest.raises(TypeError, match="missing 2 required positional arguments"):
        MgrsExtension.ext(item).apply()  # type: ignore


@pytest.mark.vcr()
def test_validate(ext_item: pystac.Item) -> None:
    assert ext_item.validate()


@pytest.mark.parametrize("field", ["latitude_band", "grid_square", "utm_zone"])
def test_get_field(ext_item: pystac.Item, field: str) -> None:
    prop = ext_item.properties[f"mgrs:{field}"]
    attr = getattr(MgrsExtension.ext(ext_item), field)

    assert attr is not None
    assert attr == prop


@pytest.mark.vcr()
@pytest.mark.parametrize(
    "field,value",
    [
        ("latitude_band", "C"),
        ("grid_square", "ZA"),
        ("utm_zone", 59),
    ],
)
def test_set_field(ext_item: pystac.Item, field: str, value) -> None:  # type: ignore
    original = ext_item.properties[f"mgrs:{field}"]
    setattr(MgrsExtension.ext(ext_item), field, value)
    new = ext_item.properties[f"mgrs:{field}"]

    assert new != original
    assert new == value
    assert ext_item.validate()


def test_utm_zone_set_to_none_pops_from_dict(ext_item: pystac.Item) -> None:
    assert "mgrs:utm_zone" in ext_item.properties

    MgrsExtension.ext(ext_item).utm_zone = None
    assert "mgrs:utm_zone" not in ext_item.properties


def test_invalid_latitude_band_raises_informative_error(ext_item: pystac.Item) -> None:
    with pytest.raises(ValueError, match="must be str"):
        MgrsExtension.ext(ext_item).latitude_band = 2  # type: ignore

    with pytest.raises(ValueError, match="must be str"):
        MgrsExtension.ext(ext_item).latitude_band = None  # type: ignore

    with pytest.raises(ValueError, match="a is not in "):
        MgrsExtension.ext(ext_item).latitude_band = "a"


def test_invalid_grid_square_raises_informative_error(ext_item: pystac.Item) -> None:
    with pytest.raises(ValueError, match="must be str"):
        MgrsExtension.ext(ext_item).grid_square = 2  # type: ignore

    with pytest.raises(ValueError, match="must be str"):
        MgrsExtension.ext(ext_item).grid_square = None  # type: ignore

    with pytest.raises(ValueError, match="nv does not match the regex "):
        MgrsExtension.ext(ext_item).grid_square = "nv"


def test_invalid_utm_zone_raises_informative_error(ext_item: pystac.Item) -> None:
    with pytest.raises(ValueError, match="must be None or int"):
        MgrsExtension.ext(ext_item).utm_zone = "foo"  # type: ignore

    with pytest.raises(ValueError, match="61 is not in "):
        MgrsExtension.ext(ext_item).utm_zone = 61



"""Tests for pystac.tests.extensions.xarray_assets"""

import json

import pytest

import pystac
from pystac.extensions.xarray_assets import XarrayAssetsExtension
from tests.conftest import get_data_file


@pytest.fixture
def ext_collection_uri() -> str:
    return get_data_file("xarray-assets/collection.json")


@pytest.fixture
def ext_collection(ext_collection_uri: str) -> pystac.Collection:
    return pystac.Collection.from_file(ext_collection_uri)


@pytest.fixture
def ext_item_uri() -> str:
    return get_data_file("xarray-assets/item.json")


@pytest.fixture
def ext_item(ext_item_uri: str) -> pystac.Item:
    return pystac.Item.from_file(ext_item_uri)


@pytest.fixture
def ext_asset(ext_item: pystac.Item) -> pystac.Asset:
    return ext_item.assets["data"]


def test_item_stac_extensions(ext_item: pystac.Item) -> None:
    assert XarrayAssetsExtension.has_extension(ext_item)


def test_collection_stac_extensions(ext_collection: pystac.Collection) -> None:
    assert XarrayAssetsExtension.has_extension(ext_collection)


def test_item_get_schema_uri(ext_item: pystac.Item) -> None:
    assert XarrayAssetsExtension.get_schema_uri() in ext_item.stac_extensions


def test_collection_get_schema_uri(ext_collection: pystac.Collection) -> None:
    assert XarrayAssetsExtension.get_schema_uri() in ext_collection.stac_extensions


def test_ext_raises_if_item_does_not_conform(item: pystac.Item) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        XarrayAssetsExtension.ext(item)


def test_ext_raises_if_collection_does_not_conform(
    collection: pystac.Collection,
) -> None:
    with pytest.raises(pystac.errors.ExtensionNotImplemented):
        XarrayAssetsExtension.ext(collection)


def test_ext_raises_on_catalog(catalog: pystac.Catalog) -> None:
    with pytest.raises(
        pystac.errors.ExtensionTypeError,
        match="XarrayAssetsExtension does not apply to type 'Catalog'",
    ):
        XarrayAssetsExtension.ext(catalog)  # type: ignore


def test_item_to_from_dict(ext_item_uri: str, ext_item: pystac.Item) -> None:
    with open(ext_item_uri) as f:
        d = json.load(f)
    actual = ext_item.to_dict(include_self_link=False)
    assert actual == d


def test_collection_to_from_dict(
    ext_collection_uri: str, ext_collection: pystac.Item
) -> None:
    with open(ext_collection_uri) as f:
        d = json.load(f)
    actual = ext_collection.to_dict(include_self_link=False)
    assert actual == d


def test_add_to_item(item: pystac.Item) -> None:
    assert not XarrayAssetsExtension.has_extension(item)
    XarrayAssetsExtension.add_to(item)

    assert XarrayAssetsExtension.has_extension(item)


def test_add_to_collection(collection: pystac.Collection) -> None:
    assert not XarrayAssetsExtension.has_extension(collection)
    XarrayAssetsExtension.add_to(collection)

    assert XarrayAssetsExtension.has_extension(collection)


@pytest.mark.vcr
def test_item_validate(ext_item: pystac.Item) -> None:
    assert ext_item.validate()


@pytest.mark.vcr
def test_collection_validate(ext_collection: pystac.Collection) -> None:
    assert ext_collection.validate()


def test_fields_are_not_on_item(ext_item: pystac.Item) -> None:
    assert not hasattr(XarrayAssetsExtension.ext(ext_item), "storage_options")
    assert not hasattr(XarrayAssetsExtension.ext(ext_item), "open_kwargs")


def test_fields_are_not_on_collection(ext_collection: pystac.Item) -> None:
    assert not hasattr(XarrayAssetsExtension.ext(ext_collection), "storage_options")
    assert not hasattr(XarrayAssetsExtension.ext(ext_collection), "open_kwargs")


@pytest.mark.parametrize("field", ["storage_options", "open_kwargs"])
def test_get_field(ext_asset: pystac.Asset, field: str) -> None:
    prop = ext_asset.extra_fields[f"xarray:{field}"]
    attr = getattr(XarrayAssetsExtension.ext(ext_asset), field)

    assert attr is not None
    assert attr == prop


@pytest.mark.parametrize(
    "field,value",
    [
        ("storage_options", {"anon": True}),
        ("open_kwargs", {"engine": "zarr"}),
    ],
)
@pytest.mark.vcr
def test_set_field(ext_asset: pystac.Asset, field: str, value) -> None:  # type: ignore
    original = ext_asset.extra_fields[f"xarray:{field}"]
    setattr(XarrayAssetsExtension.ext(ext_asset), field, value)
    new = ext_asset.extra_fields[f"xarray:{field}"]

    assert new != original
    assert new == value

    item = ext_asset.owner
    assert item is not None
    assert isinstance(item, pystac.Item)
    assert item.validate()


@pytest.mark.parametrize("field", ["storage_options", "open_kwargs"])
def test_set_field_to_none_pops_from_dict(ext_asset: pystac.Asset, field: str) -> None:
    prop_name = f"xarray:{field}"
    assert prop_name in ext_asset.extra_fields

    setattr(XarrayAssetsExtension.ext(ext_asset), field, None)
    assert prop_name not in ext_asset.extra_fields



from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import pystac
from pystac.serialization.identify import (
    OldExtensionShortIDs,
    STACJSONDescription,
    STACVersionID,
)
from pystac.version import STACVersion

if TYPE_CHECKING:
    from pystac import STACObjectType


def _migrate_catalog(
    d: dict[str, Any], version: STACVersionID, info: STACJSONDescription
) -> None:
    d["type"] = pystac.STACObjectType.CATALOG
    _migrate_license(d)


def _migrate_collection_summaries(
    d: dict[str, Any], version: STACVersionID, info: STACJSONDescription
) -> None:
    if version < "1.0.0-rc.1":
        for prop, summary in d.get("summaries", {}).items():
            if isinstance(summary, dict) and "min" in summary and "max" in summary:
                d["summaries"][prop] = {
                    "minimum": summary["min"],
                    "maximum": summary["max"],
                }


def _migrate_collection(
    d: dict[str, Any], version: STACVersionID, info: STACJSONDescription
) -> None:
    d["type"] = pystac.STACObjectType.COLLECTION
    _migrate_license(d)
    _migrate_collection_summaries(d, version, info)


def _migrate_item(
    d: dict[str, Any], version: STACVersionID, info: STACJSONDescription
) -> None:
    _migrate_license(d["properties"])


def _migrate_license(d: dict[str, Any]) -> None:
    if d.get("license") in ["various", "proprietary"]:
        d["license"] = "other"

    return None


def _migrate_datetime_range(
    d: dict[str, Any], version: STACVersionID, info: STACJSONDescription
) -> set[str] | None:
    if version < "0.9":
        # Datetime range was removed
        if (
            "dtr:start_datetime" in d["properties"]
            and "start_datetime" not in d["properties"]
        ):
            d["properties"]["start_datetime"] = d["properties"]["dtr:start_datetime"]
            del d["properties"]["dtr:start_datetime"]

        if (
            "dtr:end_datetime" in d["properties"]
            and "end_datetime" not in d["properties"]
        ):
            d["properties"]["end_datetime"] = d["properties"]["dtr:end_datetime"]
            del d["properties"]["dtr:end_datetime"]

    return None


def _get_object_migrations() -> dict[
    str, Callable[[dict[str, Any], STACVersionID, STACJSONDescription], None]
]:
    return {
        pystac.STACObjectType.CATALOG: _migrate_catalog,
        pystac.STACObjectType.COLLECTION: _migrate_collection,
        pystac.STACObjectType.ITEM: _migrate_item,
    }


def _get_removed_extension_migrations() -> dict[
    str,
    tuple[
        list[STACObjectType] | None,
        None
        | (
            Callable[
                [dict[str, Any], STACVersionID, STACJSONDescription],
                set[str] | None,
            ]
        ),
    ],
]:
    """Handles removed extensions.

    This does not handle renamed extension or extensions that were absorbed
    by other extensions; for instance the FileExtensions handles the migration of
    the previous 'checksum' extension.

    Dict of the extension ID to a tuple of optional list of STACObjectType which it was
    removed for (or None if removed from all), and an optional migrate function
    that can modify the object in case the extension was removed but the properties
    were moved.
    """
    return {
        # -- Removed in 1.0
        # assets in collections became a core property
        OldExtensionShortIDs.COLLECTION_ASSETS.value: (None, None),
        # Extensions that were placed on Collections that applied to
        # the 'commons' properties of their Items, but since the commons
        # property merging has gone away these extensions are removed
        # from the collection. Note that a migrated Collection may still
        # have a "properties" in extra_fields with the extension fields.
        OldExtensionShortIDs.EO.value: ([pystac.STACObjectType.COLLECTION], None),
        OldExtensionShortIDs.FILE.value: ([pystac.STACObjectType.COLLECTION], None),
        OldExtensionShortIDs.LABEL.value: ([pystac.STACObjectType.COLLECTION], None),
        OldExtensionShortIDs.POINTCLOUD.value: (
            [pystac.STACObjectType.COLLECTION],
            None,
        ),
        OldExtensionShortIDs.PROJECTION.value: (
            [pystac.STACObjectType.COLLECTION],
            None,
        ),
        OldExtensionShortIDs.SAR.value: ([pystac.STACObjectType.COLLECTION], None),
        OldExtensionShortIDs.SAT.value: ([pystac.STACObjectType.COLLECTION], None),
        OldExtensionShortIDs.TIMESTAMPS.value: (
            [pystac.STACObjectType.COLLECTION],
            None,
        ),
        OldExtensionShortIDs.VIEW.value: ([pystac.STACObjectType.COLLECTION], None),
        # tiled-assets was never a fully published extension;
        # remove reference to the pre-1.0 RC short ID
        OldExtensionShortIDs.TILED_ASSETS.value: (None, None),
        # Single File STAC is a removed concept; is being reworked as of
        # STAC 1.0.0-RC.3. Remove short ID from PySTAC as it's unsupported
        OldExtensionShortIDs.SINGLE_FILE_STAC.value: (None, None),
        # -- Removed in 0.9.0
        "dtr": (None, _migrate_datetime_range),
        "datetime-range": (None, _migrate_datetime_range),
        "commons": (None, None),
    }


def migrate_to_latest(
    json_dict: dict[str, Any], info: STACJSONDescription
) -> dict[str, Any]:
    """Migrates the STAC JSON to the latest version

    Args:
        json_dict : The dict of STAC JSON to identify.
        info : The info from
            :func:`~pystac.serialization.identify_stac_object` that describes
            the STAC object contained in the JSON dict.

    Returns:
        dict: A copy of the dict that is migrated to the latest version (the
        version that is pystac.version.STACVersion.DEFAULT_STAC_VERSION)
    """
    from copy import deepcopy

    result = deepcopy(json_dict)
    version = info.version_range.latest_valid_version()

    object_migrations = _get_object_migrations()
    removed_extension_migrations = _get_removed_extension_migrations()

    if version != STACVersion.DEFAULT_STAC_VERSION:
        object_migrations[info.object_type](result, version, info)
        result["stac_version"] = STACVersion.DEFAULT_STAC_VERSION

    # Ensure stac_extensions property for consistency
    result["stac_extensions"] = result.get("stac_extensions", None) or []

    pystac.EXTENSION_HOOKS.migrate(result, version, info)
    for ext in result["stac_extensions"][:]:
        if ext in removed_extension_migrations:
            object_types, migration_fn = removed_extension_migrations[ext]
            if object_types is None or info.object_type in object_types:
                if migration_fn:
                    migration_fn(result, version, info)
                result["stac_extensions"].remove(ext)

    return result



from datetime import datetime
from typing import Any

import pytest

from pystac import CommonMetadata, Item, Provider, ProviderRole, utils
from tests.utils import TestCases


@pytest.fixture
def date_time_range_item() -> Item:
    return Item.from_file(
        TestCases.get_path(
            "data-files/examples/1.0.0-beta.2/item-spec/examples/datetimerange.json"
        )
    )


@pytest.fixture
def sample_full_item() -> Item:
    return Item.from_file(
        TestCases.get_path(
            "data-files/examples/1.0.0-beta.2/item-spec/examples/sample-full.json"
        )
    )


def test_datetimes(date_time_range_item: Item) -> None:
    # save dict of original item to check that `common_metadata`
    # method doesn't mutate self.item_1
    before = date_time_range_item.clone().to_dict()
    start_datetime_str = date_time_range_item.properties["start_datetime"]
    assert isinstance(start_datetime_str, str)

    cm = date_time_range_item.common_metadata
    assert isinstance(cm, CommonMetadata)
    assert isinstance(cm.start_datetime, datetime)
    assert before == date_time_range_item.to_dict()
    assert cm.providers is None


def test_common_metadata_start_datetime(date_time_range_item: Item) -> None:
    x = date_time_range_item.clone()
    start_datetime_str = "2018-01-01T13:21:30Z"
    start_datetime_dt = utils.str_to_datetime(start_datetime_str)
    example_datetime_str = "2020-01-01T00:00:00Z"
    example_datetime_dt = utils.str_to_datetime(example_datetime_str)

    assert x.common_metadata.start_datetime == start_datetime_dt
    assert x.properties["start_datetime"] == start_datetime_str

    x.common_metadata.start_datetime = example_datetime_dt

    assert x.common_metadata.start_datetime == example_datetime_dt
    assert x.properties["start_datetime"] == example_datetime_str


def test_common_metadata_end_datetime(date_time_range_item: Item) -> None:
    x = date_time_range_item.clone()
    end_datetime_str = "2018-01-01T13:31:30Z"
    end_datetime_dt = utils.str_to_datetime(end_datetime_str)
    example_datetime_str = "2020-01-01T00:00:00Z"
    example_datetime_dt = utils.str_to_datetime(example_datetime_str)

    assert x.common_metadata.end_datetime == end_datetime_dt
    assert x.properties["end_datetime"] == end_datetime_str

    x.common_metadata.end_datetime = example_datetime_dt

    assert x.common_metadata.end_datetime == example_datetime_dt
    assert x.properties["end_datetime"] == example_datetime_str


def test_common_metadata_created(sample_full_item: Item) -> None:
    x = sample_full_item.clone()
    created_str = "2016-05-04T00:00:01Z"
    created_dt = utils.str_to_datetime(created_str)
    example_datetime_str = "2020-01-01T00:00:00Z"
    example_datetime_dt = utils.str_to_datetime(example_datetime_str)

    assert x.common_metadata.created == created_dt
    assert x.properties["created"] == created_str

    x.common_metadata.created = example_datetime_dt

    assert x.common_metadata.created == example_datetime_dt
    assert x.properties["created"] == example_datetime_str


def test_common_metadata_updated(sample_full_item: Item) -> None:
    x = sample_full_item.clone()
    updated_str = "2017-01-01T00:30:55Z"
    updated_dt = utils.str_to_datetime(updated_str)
    example_datetime_str = "2020-01-01T00:00:00Z"
    example_datetime_dt = utils.str_to_datetime(example_datetime_str)

    assert x.common_metadata.updated == updated_dt
    assert x.properties["updated"] == updated_str

    x.common_metadata.updated = example_datetime_dt

    assert x.common_metadata.updated == example_datetime_dt
    assert x.properties["updated"] == example_datetime_str


def test_common_metadata_providers(sample_full_item: Item) -> None:
    x = sample_full_item.clone()

    providers_dict_list: list[dict[str, Any]] = [
        {
            "name": "CoolSat",
            "roles": ["producer", "licensor"],
            "url": "https://cool-sat.com/",
        }
    ]
    providers_object_list = [Provider.from_dict(d) for d in providers_dict_list]

    example_providers_dict_list: list[dict[str, Any]] = [
        {
            "name": "ExampleProvider_1",
            "roles": ["example_role_1", "example_role_2"],
            "url": "https://exampleprovider1.com/",
        },
        {
            "name": "ExampleProvider_2",
            "roles": ["example_role_1", "example_role_2"],
            "url": "https://exampleprovider2.com/",
        },
    ]
    example_providers_object_list = [
        Provider.from_dict(d) for d in example_providers_dict_list
    ]

    for i in range(len(utils.get_opt(x.common_metadata.providers))):
        p1 = utils.get_opt(x.common_metadata.providers)[i]
        p2 = providers_object_list[i]
        assert isinstance(p1, Provider)
        assert isinstance(p2, Provider)
        assert p1.to_dict() == p2.to_dict()

        pd1 = x.properties["providers"][i]
        pd2 = providers_dict_list[i]
        assert isinstance(pd1, dict)
        assert isinstance(pd2, dict)
        assert pd1 == pd2

    x.common_metadata.providers = example_providers_object_list

    for i in range(len(x.common_metadata.providers)):
        p1 = x.common_metadata.providers[i]
        p2 = example_providers_object_list[i]
        assert isinstance(p1, Provider)
        assert isinstance(p2, Provider)
        assert p1.to_dict() == p2.to_dict()

        pd1 = x.properties["providers"][i]
        pd2 = example_providers_dict_list[i]
        assert isinstance(pd1, dict)
        assert isinstance(pd2, dict)
        assert pd1 == pd2


def test_common_metadata_basics(sample_full_item: Item) -> None:
    x = sample_full_item.clone()

    # Title
    title = "A CS3 item"
    example_title = "example title"
    assert x.common_metadata.title == title
    x.common_metadata.title = example_title
    assert x.common_metadata.title == example_title
    assert x.properties["title"] == example_title

    # Description
    example_description = "example description"
    assert x.common_metadata.description is None
    x.common_metadata.description = example_description
    assert x.common_metadata.description == example_description
    assert x.properties["description"] == example_description
    with pytest.raises(ValueError):
        x.common_metadata.description = ""

    # License
    license = "PDDL-1.0"
    example_license = "example license"
    assert x.common_metadata.license == license
    x.common_metadata.license = example_license
    assert x.common_metadata.license == example_license
    assert x.properties["license"] == example_license

    # Platform
    platform = "coolsat2"
    example_platform = "example_platform"
    assert x.common_metadata.platform == platform
    x.common_metadata.platform = example_platform
    assert x.common_metadata.platform == example_platform
    assert x.properties["platform"] == example_platform

    # Instruments
    instruments = ["cool_sensor_v1"]
    example_instruments = ["example instrument 1", "example instrument 2"]
    assert (x.common_metadata.instruments or []) == instruments
    x.common_metadata.instruments = example_instruments
    assert x.common_metadata.instruments == example_instruments
    assert x.properties["instruments"] == example_instruments

    # Constellation
    example_constellation = "example constellation"
    assert x.common_metadata.constellation is None
    x.common_metadata.constellation = example_constellation
    assert x.common_metadata.constellation == example_constellation
    assert x.properties["constellation"] == example_constellation

    # Mission
    example_mission = "example mission"
    assert x.common_metadata.mission is None
    x.common_metadata.mission = example_mission
    assert x.common_metadata.mission == example_mission
    assert x.properties["mission"] == example_mission

    # GSD
    gsd = 0.512
    example_gsd = 0.75
    assert x.common_metadata.gsd == gsd
    x.common_metadata.gsd = example_gsd
    assert x.common_metadata.gsd == example_gsd
    assert x.properties["gsd"] == example_gsd


class TestAssetCommonMetadata:
    @pytest.fixture
    def item(self) -> Item:
        return Item.from_file(
            TestCases.get_path("data-files/item/sample-item-asset-properties.json")
        )

    def test_title(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.title
        a2_known_value = "Thumbnail"

        # Get
        assert thumbnail_cm.title != item_value
        assert thumbnail_cm.title == a2_known_value

        # Set
        set_value = "Just Another Asset"
        analytic_cm.title = set_value

        assert analytic_cm.title == set_value
        assert analytic.to_dict()["title"] == set_value

    def test_description(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.description
        a2_known_value = "Thumbnail of the item"

        # Get
        assert thumbnail_cm.description != item_value
        assert thumbnail_cm.description == a2_known_value

        # Set
        set_value = "Yet another description."
        analytic_cm.description = set_value

        assert analytic_cm.description == set_value
        assert analytic.to_dict()["description"] == set_value

    def test_start_datetime(self, item: Item) -> None:
        item_cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = item_cm.start_datetime
        a2_known_value = utils.str_to_datetime("2017-05-01T13:22:30.040Z")

        # Get
        assert thumbnail_cm.start_datetime != item_value
        assert thumbnail_cm.start_datetime == a2_known_value

        # Set
        set_value = utils.str_to_datetime("2014-05-01T13:22:30.040Z")
        analytic_cm.start_datetime = set_value

        assert analytic_cm.start_datetime == set_value
        assert analytic.to_dict()["start_datetime"] == utils.datetime_to_str(set_value)

    def test_end_datetime(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.end_datetime
        a2_known_value = utils.str_to_datetime("2017-05-02T13:22:30.040Z")

        # Get
        assert thumbnail_cm.end_datetime != item_value
        assert thumbnail_cm.end_datetime == a2_known_value

        # Set
        set_value = utils.str_to_datetime("2014-05-01T13:22:30.040Z")
        analytic_cm.end_datetime = set_value

        assert analytic_cm.end_datetime == set_value
        assert analytic.to_dict()["end_datetime"] == utils.datetime_to_str(set_value)

    def test_license(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.license
        a2_known_value = "CC-BY-4.0"

        # Get
        assert thumbnail_cm.license != item_value
        assert thumbnail_cm.license == a2_known_value

        # Set
        set_value = "various"
        analytic_cm.license = set_value

        assert analytic_cm.license == set_value
        assert analytic.to_dict()["license"] == set_value

    def test_providers(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.providers
        a2_known_value = [
            Provider(
                name="USGS",
                url="https://landsat.usgs.gov/",
                roles=[ProviderRole.PRODUCER, ProviderRole.LICENSOR],
            )
        ]

        # Get
        assert thumbnail_cm.providers != item_value
        assert thumbnail_cm.providers == a2_known_value

        # Set
        set_value = [
            Provider(
                name="John Snow",
                url="https://cholera.com/",
                roles=[ProviderRole.PRODUCER],
            )
        ]
        analytic_cm.providers = set_value

        assert analytic_cm.providers == set_value
        assert analytic.to_dict()["providers"] == [p.to_dict() for p in set_value]

    def test_platform(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.platform
        a2_known_value = "shoes"

        # Get
        assert thumbnail_cm.platform != item_value
        assert thumbnail_cm.platform == a2_known_value

        # Set
        set_value = "brick"
        analytic_cm.platform = set_value

        assert analytic_cm.platform == set_value
        assert analytic.to_dict()["platform"] == set_value

    def test_instruments(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.instruments
        a2_known_value = ["caliper"]

        # Get
        assert thumbnail_cm.instruments != item_value
        assert thumbnail_cm.instruments == a2_known_value

        # Set
        set_value = ["horns"]
        analytic_cm.instruments = set_value

        assert analytic_cm.instruments == set_value
        assert analytic.to_dict()["instruments"] == set_value

    def test_constellation(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.constellation
        a2_known_value = "little dipper"

        # Get
        assert thumbnail_cm.constellation != item_value
        assert thumbnail_cm.constellation == a2_known_value

        # Set
        set_value = "orion"
        analytic_cm.constellation = set_value

        assert analytic_cm.constellation == set_value
        assert analytic.to_dict()["constellation"] == set_value

    def test_mission(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.mission
        a2_known_value = "possible"

        # Get
        assert thumbnail_cm.mission != item_value
        assert thumbnail_cm.mission == a2_known_value

        # Set
        set_value = "critical"
        analytic_cm.mission = set_value

        assert analytic_cm.mission == set_value
        assert analytic.to_dict()["mission"] == set_value

    def test_gsd(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.gsd
        a2_known_value = 40

        # Get
        assert thumbnail_cm.gsd != item_value
        assert thumbnail_cm.gsd == a2_known_value

        # Set
        set_value = 100
        analytic_cm.gsd = set_value

        assert analytic_cm.gsd == set_value
        assert analytic.to_dict()["gsd"] == set_value

    def test_created(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.created
        a2_known_value = utils.str_to_datetime("2017-05-17T13:22:30.040Z")

        # Get
        assert thumbnail_cm.created != item_value
        assert thumbnail_cm.created == a2_known_value

        # Set
        set_value = utils.str_to_datetime("2014-05-17T13:22:30.040Z")
        analytic_cm.created = set_value

        assert analytic_cm.created == set_value
        assert analytic.to_dict()["created"] == utils.datetime_to_str(set_value)

    def test_updated(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.updated
        a2_known_value = utils.str_to_datetime("2017-05-18T13:22:30.040Z")

        # Get
        assert thumbnail_cm.updated != item_value
        assert thumbnail_cm.updated == a2_known_value

        # Set
        set_value = utils.str_to_datetime("2014-05-18T13:22:30.040Z")
        analytic_cm.updated = set_value

        assert analytic_cm.updated == set_value
        assert analytic.to_dict()["updated"] == utils.datetime_to_str(set_value)

    def test_keywords(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.keywords
        a2_known_value = ["keyword_a"]

        # Get
        assert thumbnail_cm.keywords != item_value
        assert thumbnail_cm.keywords == a2_known_value

        # Set
        set_value = ["keyword_b"]
        analytic_cm.keywords = set_value

        assert analytic_cm.keywords == set_value
        assert analytic.to_dict()["keywords"] == set_value

    def test_roles(self, item: Item) -> None:
        cm = item.common_metadata
        analytic = item.assets["analytic"]
        analytic_cm = CommonMetadata(analytic)
        thumbnail = item.assets["thumbnail"]
        thumbnail_cm = CommonMetadata(thumbnail)

        item_value = cm.roles
        a2_known_value = ["a_role"]

        # Get
        assert thumbnail_cm.roles != item_value
        assert thumbnail_cm.roles == a2_known_value

        # Set
        set_value = ["another_role"]
        analytic_cm.roles = set_value

        assert analytic_cm.roles == set_value
        assert analytic.to_dict()["roles"] == set_value



"""Implements the :stac-ext:`Classification <classification>`."""

from __future__ import annotations

import re
import warnings
from collections.abc import Iterable
from re import Pattern
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.extensions.raster import RasterBand
from pystac.serialization.identify import STACJSONDescription, STACVersionID
from pystac.utils import get_required, map_opt

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset`,
#: :class:`~pystac.ItemAssetDefinition` or :class:`~pystac.extensions.raster.RasterBand`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition, RasterBand)

SCHEMA_URI_PATTERN: str = (
    "https://stac-extensions.github.io/classification/v{version}/schema.json"
)
DEFAULT_VERSION: str = "2.0.0"
SUPPORTED_VERSIONS: list[str] = ["2.0.0", "1.1.0", "1.0.0"]

# Field names
PREFIX: str = "classification:"
BITFIELDS_PROP: str = PREFIX + "bitfields"
CLASSES_PROP: str = PREFIX + "classes"
RASTER_BANDS_PROP: str = "raster:bands"

COLOR_HINT_PATTERN: Pattern[str] = re.compile("^([0-9A-Fa-f]{6})$")


class Classification:
    """Represents a single category of a classification.

    Use Classification.create to create a new Classification.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(
        self,
        value: int,
        description: str | None = None,
        name: str | None = None,
        title: str | None = None,
        color_hint: str | None = None,
        nodata: bool | None = None,
        percentage: float | None = None,
        count: int | None = None,
    ) -> None:
        """
        Set the properties for a new Classification.

        Args:
            value: The integer value corresponding to this class
            description: The description of this class
            name: Short name of the class for machine readability. Must consist only
                of letters, numbers, -, and _ characters. Required as of v2.0 of
                this extension.
            title: Human-readable name for use in, e.g., a map legend
            color_hint: An optional hexadecimal string-encoded representation of the
                RGB color that is suggested to represent this class (six hexadecimal
                characters, all capitalized)
            nodata: If set to true classifies a value as a no-data value.
            percentage: The percentage of data values that belong to this class
                in comparison to all data values, in percent (0-100).
            count: The number of data values that belong to this class.
        """
        self.value = value
        # TODO pystac v2.0: make `name` non-optional, move it before
        # `description` in the arg list, and remove this check
        if name is None:
            raise Exception(
                "As of v2.0.0 of the classification extension, 'name' is required"
            )
        self.name = name
        self.title = title
        self.description = description
        self.color_hint = color_hint
        self.nodata = nodata
        self.percentage = percentage
        self.count = count

        if color_hint is not None:
            match = COLOR_HINT_PATTERN.match(color_hint)
            assert (
                color_hint is None or match is not None and match.group() == color_hint
            ), "Must format color hints as '^([0-9A-F]{6})$'"

    @classmethod
    def create(
        cls,
        value: int,
        description: str | None = None,
        name: str | None = None,
        title: str | None = None,
        color_hint: str | None = None,
        nodata: bool | None = None,
        percentage: float | None = None,
        count: int | None = None,
    ) -> Classification:
        """
        Create a new Classification.

        Args:
            value: The integer value corresponding to this class
            description: The optional long-form description of this class
            name: Short name of the class for machine readability. Must consist only
                of letters, numbers, -, and _ characters. Required as of v2.0 of
                this extension.
            title: Human-readable name for use in, e.g., a map legend
            color_hint: An optional hexadecimal string-encoded representation of the
                RGB color that is suggested to represent this class (six hexadecimal
                characters, all capitalized)
            nodata: If set to true classifies a value as a no-data value.
            percentage: The percentage of data values that belong to this class
                in comparison to all data values, in percent (0-100).
            count: The number of data values that belong to this class.
        """
        c = cls({})
        c.apply(
            value=value,
            name=name,
            title=title,
            description=description,
            color_hint=color_hint,
            nodata=nodata,
            percentage=percentage,
            count=count,
        )
        return c

    @property
    def value(self) -> int:
        """Get or set the class value

        Returns:
            int
        """
        return cast(int, get_required(self.properties.get("value"), self, "value"))

    @value.setter
    def value(self, v: int) -> None:
        self.properties["value"] = v

    @property
    def description(self) -> str | None:
        """Get or set the description of the class

        Returns:
            str
        """
        return self.properties.get("description")

    @description.setter
    def description(self, v: str | None) -> None:
        if v is not None:
            self.properties["description"] = v
        else:
            self.properties.pop("description", None)

    @property
    def name(self) -> str:
        """Get or set the name of the class

        Returns:
            Optional[str]
        """
        return cast(str, get_required(self.properties.get("name"), self, "name"))

    @name.setter
    def name(self, v: str) -> None:
        if v is None:
            raise Exception(
                "`name` was converted to a required attribute in classification"
                " version v2.0, so cannot be set to None"
            )
        self.properties["name"] = v

    @property
    def title(self) -> str | None:
        return self.properties.get("title")

    @title.setter
    def title(self, v: str | None) -> None:
        if v is not None:
            self.properties["title"] = v
        else:
            self.properties.pop("title", None)

    @property
    def color_hint(self) -> str | None:
        """Get or set the optional color hint for this class.

        The color hint must be a six-character string of capitalized hexadecimal
        characters ([0-9A-F]).

        Returns:
            Optional[str]
        """
        return self.properties.get("color_hint")

    @color_hint.setter
    def color_hint(self, v: str | None) -> None:
        if v is not None:
            match = COLOR_HINT_PATTERN.match(v)
            assert v is None or match is not None and match.group() == v, (
                "Must format color hints as '^([0-9A-F]{6})$'"
            )
            self.properties["color_hint"] = v
        else:
            self.properties.pop("color_hint", None)

    @property
    def nodata(self) -> bool | None:
        """Get or set the nodata value for this class.

        Returns:
            bool | None
        """
        return self.properties.get("nodata")

    @nodata.setter
    def nodata(self, v: bool | None) -> None:
        if v is not None:
            self.properties["nodata"] = v
        else:
            self.properties.pop("nodata", None)

    @property
    def percentage(self) -> float | None:
        """Get or set the percentage value for this class.

        Returns:
            Optional[float]
        """
        return self.properties.get("percentage")

    @percentage.setter
    def percentage(self, v: float | None) -> None:
        if v is not None:
            self.properties["percentage"] = v
        else:
            self.properties.pop("percentage", None)

    @property
    def count(self) -> int | None:
        """Get or set the count value for this class.

        Returns:
            Optional[int]
        """
        return self.properties.get("count")

    @count.setter
    def count(self, v: int | None) -> None:
        if v is not None:
            self.properties["count"] = v
        else:
            self.properties.pop("count", None)

    def to_dict(self) -> dict[str, Any]:
        """Returns the dictionary encoding of this class

        Returns:
            dict: The serialization of the Classification
        """
        return self.properties

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Classification):
            raise NotImplementedError
        return (
            self.value == other.value
            and self.description == other.description
            and self.name == other.name
            and self.color_hint == other.color_hint
        )

    def __repr__(self) -> str:
        return f"<Classification value={self.value} description={self.description}>"


class Bitfield:
    """Encodes the representation of values as bits in an integer.

    Use Bitfield.create to create a new Bitfield.
    """

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]):
        self.properties = properties

    def apply(
        self,
        offset: int,
        length: int,
        classes: list[Classification],
        roles: list[str] | None = None,
        description: str | None = None,
        name: str | None = None,
    ) -> None:
        """Sets the properties for this Bitfield.

        Args:
            offset: describes the position of the least significant bit captured
                by this bitfield, with zero indicating the least significant binary
                digit
            length: the number of bits described by this bitfield
            classes: a list of Classification objects describing the various levels
                captured by this bitfield
            roles: the optional role of this bitfield (see `Asset Roles
                <https://github.com/radiantearth/stac-spec/blob/master/item-spec/item-spec.md#asset-roles>`)
            description: an optional short description of the classification
            name: the optional name of the class for machine readability
        """
        self.offset = offset
        self.length = length
        self.classes = classes
        self.roles = roles
        self.description = description
        self.name = name

        assert offset >= 0, "Non-negative offsets only"
        assert length >= 1, "Positive field lengths only"
        assert len(classes) > 0, "Must specify at least one class"
        assert roles is None or len(roles) > 0, (
            "When set, roles must contain at least one item"
        )

    @classmethod
    def create(
        cls,
        offset: int,
        length: int,
        classes: list[Classification],
        roles: list[str] | None = None,
        description: str | None = None,
        name: str | None = None,
    ) -> Bitfield:
        """Sets the properties for this Bitfield.

        Args:
            offset: describes the position of the least significant bit captured
                by this bitfield, with zero indicating the least significant binary
                digit
            length: the number of bits described by this bitfield
            classes: a list of Classification objects describing the various levels
                captured by this bitfield
            roles: the optional role of this bitfield (see `Asset Roles
                <https://github.com/radiantearth/stac-spec/blob/master/item-spec/item-spec.md#asset-roles>`)
            description: an optional short description of the classification
            name: the optional name of the class for machine readability
        """
        b = cls({})
        b.apply(
            offset=offset,
            length=length,
            classes=classes,
            roles=roles,
            description=description,
            name=name,
        )
        return b

    @property
    def offset(self) -> int:
        """Get or set the offset of the bitfield.

        Describes the position of the least significant bit captured by this
        bitfield, with zero indicating the least significant binary digit

        Returns:
            int
        """
        return cast(int, get_required(self.properties.get("offset"), self, "offset"))

    @offset.setter
    def offset(self, v: int) -> None:
        self.properties["offset"] = v

    @property
    def length(self) -> int:
        """Get or set the length (number of bits) of the bitfield

        Returns:
            int
        """
        return cast(int, get_required(self.properties.get("length"), self, "length"))

    @length.setter
    def length(self, v: int) -> None:
        self.properties["length"] = v

    @property
    def classes(self) -> list[Classification]:
        """Get or set the class definitions for the bitfield

        Returns:
            List[Classification]
        """

        return [
            Classification(d)
            for d in cast(
                list[dict[str, Any]],
                get_required(
                    self.properties.get("classes"),
                    self,
                    "classes",
                ),
            )
        ]

    @classes.setter
    def classes(self, v: list[Classification]) -> None:
        self.properties["classes"] = [c.to_dict() for c in v]

    @property
    def roles(self) -> list[str] | None:
        """Get or set the role of the bitfield.

        See `Asset Roles
        <https://github.com/radiantearth/stac-spec/blob/master/item-spec/item-spec.md#asset-roles>`

        Returns:
            Optional[List[str]]
        """
        return self.properties.get("roles")

    @roles.setter
    def roles(self, v: list[str] | None) -> None:
        if v is not None:
            self.properties["roles"] = v
        else:
            self.properties.pop("roles", None)

    @property
    def description(self) -> str | None:
        """Get or set the optional description of a bitfield.

        Returns:
            Optional[str]
        """
        return self.properties.get("description")

    @description.setter
    def description(self, v: str | None) -> None:
        if v is not None:
            self.properties["description"] = v
        else:
            self.properties.pop("description", None)

    @property
    def name(self) -> str | None:
        """Get or set the optional name of the bitfield.

        Returns:
            Optional[str]
        """
        return self.properties.get("name")

    @name.setter
    def name(self, v: str | None) -> None:
        if v is not None:
            self.properties["name"] = v
        else:
            self.properties.pop("name", None)

    def __repr__(self) -> str:
        return (
            f"<Bitfield offset={self.offset} length={self.length} "
            f"classes={self.classes}>"
        )

    def to_dict(self) -> dict[str, Any]:
        """Returns the dictionary encoding of this bitfield

        Returns:
            dict: The serialization of the Bitfield
        """
        return self.properties


class ClassificationExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of
    :class:`~pystac.Item`, :class:`~pystac.Asset`,
    :class:`~pystac.extensions.raster.RasterBand`, or
    :class:`~pystac.ItemAssetDefinition` with properties from the
    :stac-ext:`Classification Extension <classification>`.  This class is generic
    over the type of STAC object being extended.

    This class is not to be instantiated directly.  One can either directly use the
    subclass corresponding to the object you are extending, or the `ext` class
    method can be used to construct the proper class for you.
    """

    name: Literal["classification"] = "classification"
    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    def apply(
        self,
        classes: list[Classification] | None = None,
        bitfields: list[Bitfield] | None = None,
    ) -> None:
        """Applies the classification extension fields to the extended object.

        Note: one may set either the classes or bitfields objects, but not both.

        Args:
            classes: a list of
                :class:`~pystac.extensions.classification.Classification` objects
                describing the various classes in the classification
        """
        assert (
            classes is None
            and bitfields is not None
            or bitfields is None
            and classes is not None
        ), "Must set exactly one of `classes` or `bitfields`"
        if classes:
            self.classes = classes
        if bitfields:
            self.bitfields = bitfields

    @property
    def classes(self) -> list[Classification] | None:
        """Get or set the classes for the base object

        Note: Setting the classes will clear the object's bitfields if they are
        not None

        Returns:
            Optional[List[Classification]]
        """
        return self._get_classes()

    @classes.setter
    def classes(self, v: list[Classification] | None) -> None:
        if self._get_bitfields() is not None:
            self.bitfields = None
        self._set_property(
            CLASSES_PROP, map_opt(lambda classes: [c.to_dict() for c in classes], v)
        )

    def _get_classes(self) -> list[Classification] | None:
        return map_opt(
            lambda classes: [Classification(c) for c in classes],
            self._get_property(CLASSES_PROP, list[dict[str, Any]]),
        )

    @property
    def bitfields(self) -> list[Bitfield] | None:
        """Get or set the bitfields for the base object

        Note: Setting the bitfields will clear the object's classes if they are
        not None

        Returns:
            Optional[List[Bitfield]]
        """
        return self._get_bitfields()

    @bitfields.setter
    def bitfields(self, v: list[Bitfield] | None) -> None:
        if self._get_classes() is not None:
            self.classes = None
        self._set_property(
            BITFIELDS_PROP,
            map_opt(lambda bitfields: [b.to_dict() for b in bitfields], v),
        )

    def _get_bitfields(self) -> list[Bitfield] | None:
        return map_opt(
            lambda bitfields: [Bitfield(b) for b in bitfields],
            self._get_property(BITFIELDS_PROP, list[dict[str, Any]]),
        )

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)

    @classmethod
    def get_schema_uris(cls) -> list[str]:
        warnings.warn(
            "get_schema_uris is deprecated and will be removed in v2",
            DeprecationWarning,
        )
        return [SCHEMA_URI_PATTERN.format(version=v) for v in SUPPORTED_VERSIONS]

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> ClassificationExtension[T]:
        """Extends the given STAC object with propertied from the
        :stac-ext:`Classification Extension <classification>`

        This extension can be applied to instances of :class:`~pystac.Item`,
        :class:`~pystac.Asset`,
        :class:`~pystac.ItemAssetDefinition`, or
        :class:`~pystac.extensions.raster.RasterBand`.

        Raises:
            pystac.ExtensionTypeError : If an invalid object type is passed
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(ClassificationExtension[T], ItemClassificationExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(ClassificationExtension[T], AssetClassificationExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(
                ClassificationExtension[T], ItemAssetsClassificationExtension(obj)
            )
        elif isinstance(obj, RasterBand):
            return cast(
                ClassificationExtension[T], RasterBandClassificationExtension(obj)
            )
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesClassificationExtension:
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesClassificationExtension(obj)


class ItemClassificationExtension(ClassificationExtension[pystac.Item]):
    item: pystac.Item

    properties: dict[str, Any]

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemClassificationExtension Item id={self.item.id}>"


class AssetClassificationExtension(ClassificationExtension[pystac.Asset]):
    asset: pystac.Asset
    asset_href: str
    properties: dict[str, Any]
    additional_read_properties: Iterable[dict[str, Any]] | None

    def __init__(self, asset: pystac.Asset):
        self.asset = asset
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetClassificationExtension Asset href={self.asset_href}>"


class ItemAssetsClassificationExtension(
    ClassificationExtension[pystac.ItemAssetDefinition]
):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties

    def __repr__(self) -> str:
        return (
            f"<ItemAssetsClassificationExtension ItemAssetDefinition={self.asset_defn}"
        )


class RasterBandClassificationExtension(ClassificationExtension[RasterBand]):
    properties: dict[str, Any]
    raster_band: RasterBand

    def __init__(self, raster_band: RasterBand):
        self.raster_band = raster_band
        self.properties = raster_band.properties

    def __repr__(self) -> str:
        return f"<RasterBandClassificationExtension RasterBand={self.raster_band}>"


class SummariesClassificationExtension(SummariesExtension):
    @property
    def classes(self) -> list[Classification] | None:
        return map_opt(
            lambda classes: [Classification(c) for c in classes],
            self.summaries.get_list(CLASSES_PROP),
        )

    @classes.setter
    def classes(self, v: list[Classification] | None) -> None:
        self._set_summary(CLASSES_PROP, map_opt(lambda x: [c.to_dict() for c in x], v))

    @property
    def bitfields(self) -> list[Bitfield] | None:
        return map_opt(
            lambda bitfields: [Bitfield(b) for b in bitfields],
            self.summaries.get_list(BITFIELDS_PROP),
        )

    @bitfields.setter
    def bitfields(self, v: list[Bitfield] | None) -> None:
        self._set_summary(
            BITFIELDS_PROP, map_opt(lambda x: [b.to_dict() for b in x], v)
        )


class ClassificationExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI_PATTERN.format(version=DEFAULT_VERSION)
    prev_extension_ids = {
        SCHEMA_URI_PATTERN.format(version=v)
        for v in SUPPORTED_VERSIONS
        if v != DEFAULT_VERSION
    }
    stac_object_types = {pystac.STACObjectType.ITEM}

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        if SCHEMA_URI_PATTERN.format(version="1.0.0") in info.extensions:
            for asset in obj.get("assets", {}).values():
                classification_classes = asset.get(CLASSES_PROP, None)
                if classification_classes is None or not isinstance(
                    classification_classes, list
                ):
                    continue
                for class_object in classification_classes:
                    if "color-hint" in class_object:
                        class_object["color_hint"] = class_object["color-hint"]
                        del class_object["color-hint"]
        super().migrate(obj, version, info)


CLASSIFICATION_EXTENSION_HOOKS: ExtensionHooks = ClassificationExtensionHooks()



import json
import os
import shutil
import tempfile
from datetime import datetime
from pathlib import Path
from tempfile import TemporaryDirectory

from pystac import (
    Catalog,
    Collection,
    Extent,
    Item,
    SpatialExtent,
    StacIO,
    TemporalExtent,
)

from ._base import Bench
from ._util import get_data_path


class CatalogBench(Bench):
    def setup(self) -> None:
        self.temp_dir = tempfile.mkdtemp()

        self.stac_io = StacIO.default()

        self.catalog_path = get_data_path("examples/1.0.0/catalog.json")
        with open(self.catalog_path) as src:
            self.catalog_dict = json.load(src)
        self.catalog = Catalog.from_file(self.catalog_path)

    def teardown(self) -> None:
        shutil.rmtree(self.temp_dir, ignore_errors=True)

    def time_catalog_from_file(self) -> None:
        """Deserialize an Item from file"""
        _ = Catalog.from_file(self.catalog_path)

    def time_catalog_from_dict(self) -> None:
        """Deserialize an Item from dictionary."""
        _ = Catalog.from_dict(self.catalog_dict)

    def time_catalog_to_dict(self) -> None:
        """Serialize an Item to a dictionary."""
        self.catalog.to_dict(include_self_link=True)

    def time_catalog_save(self) -> None:
        """Serialize an Item to a JSON file."""
        self.catalog.save_object(
            include_self_link=True,
            dest_href=os.path.join(self.temp_dir, "time_catalog_save.json"),
            stac_io=self.stac_io,
        )


class WalkCatalogBench(Bench):
    def setup_cache(self) -> Catalog:
        return make_large_catalog()

    def time_walk(self, catalog: Catalog) -> None:
        for (
            _,
            _,
            _,
        ) in catalog.walk():
            pass

    def peakmem_walk(self, catalog: Catalog) -> None:
        for (
            _,
            _,
            _,
        ) in catalog.walk():
            pass


class ReadCatalogBench(Bench):
    def setup(self) -> None:
        catalog = make_large_catalog()
        self.temporary_directory = TemporaryDirectory()
        self.path = str(Path(self.temporary_directory.name) / "catalog.json")
        catalog.normalize_and_save(self.temporary_directory.name)

    def teardown(self) -> None:
        shutil.rmtree(self.temporary_directory.name)

    def time_read_and_walk(self) -> None:
        catalog = Catalog.from_file(self.path)
        for _, _, _ in catalog.walk():
            pass


class WriteCatalogBench(Bench):
    def setup(self) -> None:
        self.catalog = make_large_catalog()
        self.temporary_directory = TemporaryDirectory()

    def teardown(self) -> None:
        shutil.rmtree(self.temporary_directory.name)

    def time_normalize_and_save(self) -> None:
        self.catalog.normalize_and_save(self.temporary_directory.name)


def make_large_catalog() -> Catalog:
    catalog = Catalog("an-id", "a description")
    extent = Extent(
        SpatialExtent([[-180.0, -90.0, 180.0, 90.0]]),
        TemporalExtent([[datetime(2023, 1, 1), None]]),
    )
    for i in range(0, 10):
        collection = Collection(f"collection-{i}", f"Collection {i}", extent)
        for j in range(0, 100):
            item = Item(f"item-{i}-{j}", None, None, datetime.now(), {})
            collection.add_item(item)
        catalog.add_child(collection)
    return catalog



"""Implements the :stac-ext:`Satellite Extension <sat>`."""

from __future__ import annotations

from collections.abc import Iterable
from datetime import datetime
from typing import (
    Any,
    Generic,
    Literal,
    TypeVar,
    cast,
)

import pystac
from pystac.extensions.base import (
    ExtensionManagementMixin,
    PropertiesExtension,
    SummariesExtension,
)
from pystac.extensions.hooks import ExtensionHooks
from pystac.summaries import RangeSummary
from pystac.utils import StringEnum, datetime_to_str, map_opt, str_to_datetime

#: Generalized version of :class:`~pystac.Item`, :class:`~pystac.Asset` or
#: :class:`~pystac.ItemAssetDefinition`
T = TypeVar("T", pystac.Item, pystac.Asset, pystac.ItemAssetDefinition)

SCHEMA_URI = "https://stac-extensions.github.io/sat/v1.0.0/schema.json"

PREFIX: str = "sat:"
PLATFORM_INTERNATIONAL_DESIGNATOR_PROP: str = (
    PREFIX + "platform_international_designator"
)
ABSOLUTE_ORBIT_PROP: str = PREFIX + "absolute_orbit"
ORBIT_STATE_PROP: str = PREFIX + "orbit_state"
RELATIVE_ORBIT_PROP: str = PREFIX + "relative_orbit"
ANX_DATETIME_PROP: str = PREFIX + "anx_datetime"


class OrbitState(StringEnum):
    ASCENDING = "ascending"
    DESCENDING = "descending"
    GEOSTATIONARY = "geostationary"


class SatExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[pystac.Item | pystac.Collection],
):
    """An abstract class that can be used to extend the properties of an
    :class:`~pystac.Item` or :class:`~pystac.Asset` with properties from the
    :stac-ext:`Satellite Extension <sat>`. This class is generic over the type of
    STAC Object to be extended (e.g. :class:`~pystac.Item`,
    :class:`~pystac.Collection`).

    To create a concrete instance of :class:`SatExtension`, use the
    :meth:`SatExtension.ext` method. For example:

    .. code-block:: python

       >>> item: pystac.Item = ...
       >>> sat_ext = SatExtension.ext(item)
    """

    name: Literal["sat"] = "sat"

    def apply(
        self,
        orbit_state: OrbitState | None = None,
        relative_orbit: int | None = None,
        absolute_orbit: int | None = None,
        platform_international_designator: str | None = None,
        anx_datetime: datetime | None = None,
    ) -> None:
        """Applies ext extension properties to the extended :class:`~pystac.Item` or
        class:`~pystac.Asset`.

        Must specify at least one of orbit_state or relative_orbit in order
        for the sat extension to properties to be valid.

        Args:
            orbit_state : Optional state of the orbit. Either ascending or
                descending for polar orbiting satellites, or geostationary for
                geosynchronous satellites.
            relative_orbit : Optional non-negative integer of the orbit number at
                the time of acquisition.
        """

        self.platform_international_designator = platform_international_designator
        self.orbit_state = orbit_state
        self.absolute_orbit = absolute_orbit
        self.relative_orbit = relative_orbit
        self.anx_datetime = anx_datetime

    @property
    def platform_international_designator(self) -> str | None:
        """Gets or sets the International Designator, also known as COSPAR ID, and
        NSSDCA ID."""
        return self._get_property(PLATFORM_INTERNATIONAL_DESIGNATOR_PROP, str)

    @platform_international_designator.setter
    def platform_international_designator(self, v: str | None) -> None:
        self._set_property(PLATFORM_INTERNATIONAL_DESIGNATOR_PROP, v)

    @property
    def orbit_state(self) -> OrbitState | None:
        """Get or sets an orbit state of the object."""
        return map_opt(
            lambda x: OrbitState(x), self._get_property(ORBIT_STATE_PROP, str)
        )

    @orbit_state.setter
    def orbit_state(self, v: OrbitState | None) -> None:
        self._set_property(ORBIT_STATE_PROP, map_opt(lambda x: x.value, v))

    @property
    def absolute_orbit(self) -> int | None:
        """Get or sets an absolute orbit number of the item."""
        return self._get_property(ABSOLUTE_ORBIT_PROP, int)

    @absolute_orbit.setter
    def absolute_orbit(self, v: int | None) -> None:
        self._set_property(ABSOLUTE_ORBIT_PROP, v)

    @property
    def relative_orbit(self) -> int | None:
        """Get or sets a relative orbit number of the item."""
        return self._get_property(RELATIVE_ORBIT_PROP, int)

    @relative_orbit.setter
    def relative_orbit(self, v: int | None) -> None:
        self._set_property(RELATIVE_ORBIT_PROP, v)

    @property
    def anx_datetime(self) -> datetime | None:
        return map_opt(str_to_datetime, self._get_property(ANX_DATETIME_PROP, str))

    @anx_datetime.setter
    def anx_datetime(self, v: datetime | None) -> None:
        self._set_property(ANX_DATETIME_PROP, map_opt(datetime_to_str, v))

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: T, add_if_missing: bool = False) -> SatExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`Satellite
        Extension <sat>`.

        This extension can be applied to instances of :class:`~pystac.Item` or
        :class:`~pystac.Asset`.

        Raises:

            pystac.ExtensionTypeError : If an invalid object type is passed.
        """
        if isinstance(obj, pystac.Item):
            cls.ensure_has_extension(obj, add_if_missing)
            return cast(SatExtension[T], ItemSatExtension(obj))
        elif isinstance(obj, pystac.Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(SatExtension[T], AssetSatExtension(obj))
        elif isinstance(obj, pystac.ItemAssetDefinition):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(SatExtension[T], ItemAssetsSatExtension(obj))
        else:
            raise pystac.ExtensionTypeError(cls._ext_error_message(obj))

    @classmethod
    def summaries(
        cls, obj: pystac.Collection, add_if_missing: bool = False
    ) -> SummariesSatExtension:
        """Returns the extended summaries object for the given collection."""
        cls.ensure_has_extension(obj, add_if_missing)
        return SummariesSatExtension(obj)


class ItemSatExtension(SatExtension[pystac.Item]):
    """A concrete implementation of :class:`SatExtension` on an :class:`~pystac.Item`
    that extends the properties of the Item to include properties defined in the
    :stac-ext:`Satellite Extension <sat>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`SatExtension.ext` on an :class:`~pystac.Item` to
    extend it.
    """

    item: pystac.Item
    """The :class:`~pystac.Item` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Item` properties, including extension properties."""

    def __init__(self, item: pystac.Item):
        self.item = item
        self.properties = item.properties

    def __repr__(self) -> str:
        return f"<ItemSatExtension Item id={self.item.id}>"


class AssetSatExtension(SatExtension[pystac.Asset]):
    """A concrete implementation of :class:`SatExtension` on an :class:`~pystac.Asset`
    that extends the properties of the Asset to include properties defined in the
    :stac-ext:`Satellite Extension <sat>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`SatExtension.ext` on an :class:`~pystac.Asset` to
    extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owning :class:`~pystac.Item`."""

    def __init__(self, asset: pystac.Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and isinstance(asset.owner, pystac.Item):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetSatExtension Asset href={self.asset_href}>"


class ItemAssetsSatExtension(SatExtension[pystac.ItemAssetDefinition]):
    properties: dict[str, Any]
    asset_defn: pystac.ItemAssetDefinition

    def __init__(self, item_asset: pystac.ItemAssetDefinition):
        self.asset_defn = item_asset
        self.properties = item_asset.properties


class SummariesSatExtension(SummariesExtension):
    """A concrete implementation of :class:`~pystac.extensions.base.SummariesExtension`
    that extends the ``summaries`` field of a :class:`~pystac.Collection` to include
    properties defined in the :stac-ext:`Satellite Extension <sat>`.
    """

    @property
    def platform_international_designator(self) -> list[str] | None:
        """Get or sets the summary of
        :attr:`SatExtension.platform_international_designator` values for this
        Collection.
        """

        return self.summaries.get_list(PLATFORM_INTERNATIONAL_DESIGNATOR_PROP)

    @platform_international_designator.setter
    def platform_international_designator(self, v: list[str] | None) -> None:
        self._set_summary(PLATFORM_INTERNATIONAL_DESIGNATOR_PROP, v)

    @property
    def orbit_state(self) -> list[OrbitState] | None:
        """Get or sets the summary of :attr:`SatExtension.orbit_state` values
        for this Collection.
        """

        return self.summaries.get_list(ORBIT_STATE_PROP)

    @orbit_state.setter
    def orbit_state(self, v: list[OrbitState] | None) -> None:
        self._set_summary(ORBIT_STATE_PROP, v)

    @property
    def absolute_orbit(self) -> RangeSummary[int] | None:
        return self.summaries.get_range(ABSOLUTE_ORBIT_PROP)

    @absolute_orbit.setter
    def absolute_orbit(self, v: RangeSummary[int] | None) -> None:
        self._set_summary(ABSOLUTE_ORBIT_PROP, v)

    @property
    def relative_orbit(self) -> RangeSummary[int] | None:
        return self.summaries.get_range(RELATIVE_ORBIT_PROP)

    @relative_orbit.setter
    def relative_orbit(self, v: RangeSummary[int] | None) -> None:
        self._set_summary(RELATIVE_ORBIT_PROP, v)

    @property
    def anx_datetime(self) -> RangeSummary[datetime] | None:
        return map_opt(
            lambda s: RangeSummary(
                str_to_datetime(s.minimum), str_to_datetime(s.maximum)
            ),
            self.summaries.get_range(ANX_DATETIME_PROP),
        )

    @anx_datetime.setter
    def anx_datetime(self, v: RangeSummary[datetime] | None) -> None:
        self._set_summary(
            ANX_DATETIME_PROP,
            map_opt(
                lambda s: RangeSummary(
                    datetime_to_str(s.minimum), datetime_to_str(s.maximum)
                ),
                v,
            ),
        )


class SatExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {"sat"}
    stac_object_types = {pystac.STACObjectType.ITEM}


SAT_EXTENSION_HOOKS: ExtensionHooks = SatExtensionHooks()



from __future__ import annotations

from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any, cast

import pystac
from pystac.serialization.identify import STACVersionID, identify_stac_object
from pystac.stac_object import STACObjectType
from pystac.utils import make_absolute_href
from pystac.validation.schema_uri_map import OldExtensionSchemaUriMap
from pystac.validation.stac_validator import GetSchemaError

if TYPE_CHECKING:
    from pystac.stac_object import STACObject


# Import after above class definition
from pystac.validation.stac_validator import JsonSchemaSTACValidator, STACValidator

__all__ = [
    "GetSchemaError",
    "JsonSchemaSTACValidator",
    "RegisteredValidator",
    "validate",
    "validate_all",
    "validate_dict",
    "validate_all_dict",
    "set_validator",
]


def validate(
    stac_object: STACObject,
    validator: STACValidator | None = None,
) -> list[Any]:
    """Validates a :class:`~pystac.STACObject`.

    Args:
        stac_object : The stac object to validate.
        validator : A custom validator to use for validation of the STAC object.
            If omitted, the default validator from
            :class:`~pystac.validation.RegisteredValidator`
            will be used instead.

    Returns:
        List[Any]: List of return values from the validation calls for the
           core object and any extensions. Element type is specific to the
           STACValidator implementation.

    Raises:
        STACValidationError
    """
    return validate_dict(
        stac_dict=stac_object.to_dict(),
        stac_object_type=stac_object.STAC_OBJECT_TYPE,
        stac_version=pystac.get_stac_version(),
        extensions=stac_object.stac_extensions,
        href=stac_object.get_self_href(),
        validator=validator,
    )


def validate_dict(
    stac_dict: dict[str, Any],
    stac_object_type: STACObjectType | None = None,
    stac_version: str | None = None,
    extensions: list[str] | None = None,
    href: str | None = None,
    validator: STACValidator | None = None,
) -> list[Any]:
    """Validate a stac object serialized as JSON into a dict.

    This method delegates to the call to
    :meth:`~pystac.validation.stac_validator.STACValidator.validate` for the
    :class:`~pystac.validation.stac_validator.STACValidator` registered
    via :func:`~pystac.validation.set_validator` or
    :class:`~pystac.validation.JsonSchemaSTACValidator` by default.

    Args:
        stac_dict : Dictionary that is the STAC JSON of the object.
        stac_object_type : The stac object type of the object encoded in stac_dict.
            One of :class:`~pystac.STACObjectType`. If not supplied, this will use
            PySTAC's identification logic to identify the object type.
        stac_version : The version of STAC to validate the object against. If not
            supplied, this will use PySTAC's identification logic to identify the stac
            version
        extensions : Extension IDs for this stac object. If not supplied,
            PySTAC's identification logic to identify the extensions.
        href : Optional HREF of the STAC object being validated.
        validator : A custom validator to use for validation of the STAC dictionary.
            If omitted, the default validator from
            :class:`~pystac.validation.RegisteredValidator`
            will be used instead.

    Returns:
        List[Any]: List of return values from the validation calls for the
           core object and any extensions. Element type is specific to the
           STACValidator implementation.

    Raises:
        STACValidationError
    """
    info = None
    if stac_object_type is None:
        info = identify_stac_object(stac_dict)
        stac_object_type = info.object_type
    if stac_version is None:
        if info is None:
            info = identify_stac_object(stac_dict)
        stac_version = str(info.version_range.latest_valid_version())
    if extensions is None:
        if info is None:
            info = identify_stac_object(stac_dict)
        extensions = list(info.extensions)

    stac_version_id = STACVersionID(stac_version)

    # If the version is before 1.0.0-rc.2, substitute extension short IDs for
    # their schemas.
    if stac_version_id < "1.0.0-rc.2":

        def _get_uri(ext: str) -> str | None:
            return OldExtensionSchemaUriMap.get_extension_schema_uri(
                ext,
                stac_object_type,
                stac_version_id,
            )

        extensions = [uri for uri in map(_get_uri, extensions) if uri is not None]

    validator = validator or RegisteredValidator.get_validator()
    return validator.validate(
        stac_dict, stac_object_type, stac_version, extensions, href
    )


def validate_all(
    stac_object: STACObject | dict[str, Any],
    href: str | None = None,
    stac_io: pystac.StacIO | None = None,
) -> None:
    """Validate a :class:`~pystac.STACObject`, or a STAC object serialized as
    JSON into a dict.

    If the STAC object represents a catalog or collection, this function will be
    called recursively for each child link and all contained items.

    Args:
        stac_object : STAC object to validate
        href : HREF of the STAC object being validated. Used for error
            reporting and resolving relative links.
        stac_io : Optional StacIO instance to use for reading hrefs. If None,
            the StacIO.default() instance is used.

    Raises:
        STACValidationError or ValueError: STACValidationError is raised
            if the STAC object or any contained catalog, collection,
            or item has a validation error. ValueError is raised
            if stac_object is a STACObject and href is not None, or if
            stac_object is a dict and href is None.
    """
    if stac_io is None:
        stac_io = pystac.StacIO.default()

    if isinstance(stac_object, dict):
        import warnings

        warnings.warn(
            "validating a STAC object as a dict is deprecated;"
            " use validate_all_dict instead",
            DeprecationWarning,
        )

        if href is None:
            raise ValueError(
                "href must be set if stac_object is a dict; it will be removed"
                " once support for validating a STAC object as a dict is removed from"
                " validate_all, due to the introduction of validate_all_dict"
            )

        validate_all_dict(stac_object, href, stac_io)
    elif href is not None:
        raise ValueError(
            "href must be None if stac_object is a STACObject; it will be removed"
            " once support for validating a STAC object as a dict is removed from"
            " validate_all, due to the introduction of validate_all_dict"
        )
    else:
        validate_all_dict(stac_object.to_dict(), stac_object.get_self_href(), stac_io)


def validate_all_dict(
    stac_dict: dict[str, Any],
    href: str | None,
    stac_io: pystac.StacIO | None = None,
) -> None:
    """Validate a stac object serialized as JSON into a dict.

    If the STAC object represents a catalog or collection, this function will be
    called recursively for each child link and all contained items.

    Args:
        stac_dict : Dictionary that is the STAC JSON of the object.
        href : HREF of the STAC object being validated. Used for error
            reporting and resolving relative links.
        stac_io : Optional StacIO instance to use for reading hrefs. If None,
            the StacIO.default() instance is used.

    Raises:
        STACValidationError: if the STAC object or any contained catalog, collection,
            or item has a validation error
    """
    if stac_io is None:
        stac_io = pystac.StacIO.default()

    info = identify_stac_object(stac_dict)

    # Validate this object
    validate_dict(
        stac_dict,
        stac_object_type=info.object_type,
        stac_version=str(info.version_range.latest_valid_version()),
        extensions=list(info.extensions),
        href=href,
    )

    if info.object_type != pystac.STACObjectType.ITEM and "links" in stac_dict:
        links = (
            # Account for 0.6 links
            stac_dict["links"].values()
            if isinstance(stac_dict["links"], dict)
            else stac_dict.get("links")
        )

        for link in cast(Iterable[Mapping[str, Any]], links):
            if link.get("rel") in [pystac.RelType.ITEM, pystac.RelType.CHILD]:
                link_href = make_absolute_href(
                    cast(str, link.get("href")), start_href=href
                )
                validate_all_dict(stac_io.read_json(link_href), link_href, stac_io)


class RegisteredValidator:
    _validator: STACValidator | None = None

    @classmethod
    def get_validator(cls) -> STACValidator:
        if cls._validator is None:
            try:
                import jsonschema as _  # noqa
            except ImportError:
                raise Exception(
                    "Cannot validate with default validator because package"
                    ' "jsonschema" is not installed. Install pystac with the validation'
                    " optional requirements (e.g. pip install pystac[validation]) to"
                    " install jsonschema"
                )
            cls._validator = JsonSchemaSTACValidator()

        return cls._validator

    @classmethod
    def set_validator(cls, validator: STACValidator) -> None:
        if not issubclass(type(validator), STACValidator):
            raise Exception(f"Validator must be a subclass of {STACValidator}")
        cls._validator = validator


def set_validator(validator: STACValidator) -> None:
    """Sets the STACValidator to use in PySTAC.

    Args:
        validator : The STACValidator implementation to use for
            validation.
    """
    RegisteredValidator.set_validator(validator)



from __future__ import annotations

from enum import Enum
from functools import total_ordering
from typing import TYPE_CHECKING, Any

import pystac
from pystac.version import STACVersion

if TYPE_CHECKING:
    from pystac.stac_object import STACObjectType


class OldExtensionShortIDs(Enum):
    """Enumerates the IDs of common extensions."""

    CHECKSUM = "checksum"
    COLLECTION_ASSETS = "collection-assets"
    DATACUBE = "datacube"  # TODO
    EO = "eo"
    ITEM_ASSETS = "item-assets"  # TODO
    LABEL = "label"
    POINTCLOUD = "pointcloud"
    PROJECTION = "projection"
    SAR = "sar"
    SAT = "sat"
    SCIENTIFIC = "scientific"
    SINGLE_FILE_STAC = "single-file-stac"
    TILED_ASSETS = "tiled-assets"
    TIMESTAMPS = "timestamps"
    VERSION = "version"
    VIEW = "view"
    FILE = "file"


@total_ordering
class STACVersionID:
    """Defines STAC versions in an object that is orderable based on version number.
    For instance, ``1.0.0-beta.2 < 1.0.0``
    """

    version_string: str
    version_core: str
    version_prerelease: str | None

    def __init__(self, version_string: str) -> None:
        self.version_string = version_string

        # Account for RC or beta releases in version
        version_parts = version_string.split("-")
        self.version_core = version_parts[0]
        if len(version_parts) == 1:
            self.version_prerelease = None
        else:
            self.version_prerelease = "-".join(version_parts[1:])

    def __str__(self) -> str:
        return self.version_string

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, STACVersionID):
            other = STACVersionID(str(other))
        return str(self) == str(other)

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __lt__(self, other: object) -> bool:
        if not isinstance(other, STACVersionID):
            other = STACVersionID(str(other))
        if self.version_core < other.version_core:
            return True
        elif self.version_core > other.version_core:
            return False
        else:
            return self.version_prerelease is not None and (
                other.version_prerelease is None
                or other.version_prerelease > self.version_prerelease
            )


class STACVersionRange:
    """Defines a range of STAC versions."""

    _min_version: STACVersionID
    _max_version: STACVersionID

    def __init__(
        self,
        min_version: str | STACVersionID = "0.8.0",
        max_version: str | STACVersionID | None = None,
    ):
        if isinstance(min_version, str):
            self.min_version = STACVersionID(min_version)
        else:
            self.min_version = min_version

        if max_version is None:
            self.max_version = STACVersionID(STACVersion.DEFAULT_STAC_VERSION)
        else:
            if isinstance(max_version, str):
                self.max_version = STACVersionID(max_version)
            else:
                self.max_version = max_version

    @property
    def min_version(self) -> STACVersionID:
        return self._min_version

    @min_version.setter
    def min_version(self, v: str | STACVersionID) -> None:
        if isinstance(v, str):
            v = STACVersionID(v)
        self._min_version = v

    @property
    def max_version(self) -> STACVersionID:
        return self._max_version

    @max_version.setter
    def max_version(self, v: str | STACVersionID) -> None:
        if isinstance(v, str):
            v = STACVersionID(v)
        self._max_version = v

    def set_min(self, v: str | STACVersionID) -> None:
        if isinstance(v, str):
            v = STACVersionID(v)
        if self.min_version < v:
            if v < self.max_version:
                self.min_version = v
            else:
                self.min_version = self.max_version

    def set_max(self, v: str | STACVersionID) -> None:
        if isinstance(v, str):
            v = STACVersionID(v)
        if v < self.max_version:
            if self.min_version < v:
                self.max_version = v
            else:
                self.max_version = self.min_version

    def set_to_single(self, v: str | STACVersionID) -> None:
        self.set_min(v)
        self.set_max(v)

    def latest_valid_version(self) -> STACVersionID:
        return self.max_version

    def contains(self, v: str | STACVersionID) -> bool:
        if isinstance(v, str):
            v = STACVersionID(v)
        return self.min_version <= v <= self.max_version

    def is_single_version(self) -> bool:
        return self.min_version >= self.max_version

    def is_earlier_than(self, v: str | STACVersionID) -> bool:
        if isinstance(v, str):
            v = STACVersionID(v)
        return self.max_version < v

    def is_later_than(self, v: str | STACVersionID) -> bool:
        if isinstance(v, str):
            v = STACVersionID(v)
        return v < self.min_version

    def __repr__(self) -> str:
        return f"<VERSIONS {self.min_version}-{self.max_version}>"


class STACJSONDescription:
    """Describes the STAC object information for a STAC object represented in JSON

    Args:
        object_type : Describes the STAC object type. One of
            :class:`~pystac.STACObjectType`.
        version_range : The STAC version range that describes what
            has been identified as potential valid versions of the stac object.
        extensions : List of extension schema URIs for extensions this
            object implements
    """

    object_type: STACObjectType
    """Describes the STAC object type. One of :class:`~pystac.STACObjectType`."""

    version_range: STACVersionRange
    """The STAC version range that describes what has been identified
    as potential valid versions of the :class:`~pystac.STACObject`."""

    extensions: list[str]
    """List of extension schema URIs for extensions this object implements"""

    def __init__(
        self,
        object_type: STACObjectType,
        version_range: STACVersionRange,
        extensions: list[str],
    ) -> None:
        self.object_type = object_type
        self.version_range = version_range
        self.extensions = extensions

    def __repr__(self) -> str:
        return "<{} {} ext={}>".format(
            self.object_type, self.version_range, ",".join(self.extensions)
        )


def identify_stac_object_type(json_dict: dict[str, Any]) -> STACObjectType | None:
    """Determines the STACObjectType of the provided JSON dict. If the JSON dict does
    not represent a STAC object, returns ``None``.

    Will first try to identify the object using ``"type"`` field as described in the
    guidelines in :stac-spec:`How to Differentiate STAC Files
    <best-practices.md#how-to-differentiate-stac-files>`. If this fails, will fall back
    to using the pre-1.0 heuristic described in `this issue
    <https://github.com/radiantearth/stac-spec/issues/889#issuecomment-684529444>`__

    Args:
        json_dict : The dict of JSON to identify.
    """
    stac_version = (
        STACVersionID(json_dict["stac_version"])
        if "stac_version" in json_dict
        else None
    )
    obj_type = json_dict.get("type")

    # Try to identify using 'type' property for v1.0.0-rc.1 and higher
    introduced_type_attribute = STACVersionID("1.0.0-rc.1")
    if stac_version is not None and stac_version >= introduced_type_attribute:
        # Since v1.0.0-rc.1 requires a "type" field for all STAC objects, any object
        # that is missing this attribute is not a valid STAC object.
        if obj_type is None:
            return None

        # Try to match the "type" attribute
        if obj_type == pystac.STACObjectType.CATALOG:
            return pystac.STACObjectType.CATALOG
        elif obj_type == pystac.STACObjectType.COLLECTION:
            return pystac.STACObjectType.COLLECTION
        elif obj_type == pystac.STACObjectType.ITEM:
            return pystac.STACObjectType.ITEM
        else:
            return None

    # For pre-1.0 objects for version 0.8.* or later 'stac_version' must be present
    if stac_version is not None:
        # Pre-1.0 STAC objects with 'type' == "Feature" are Items
        if obj_type == "Feature":
            return pystac.STACObjectType.ITEM
        # Anything else with a 'type' field is not a STAC object
        if obj_type is not None:
            return None

        # Collections will contain either an 'extent' or a 'license' (or both)
        if "extent" in json_dict or "license" in json_dict:
            return pystac.STACObjectType.COLLECTION
        # Everything else that has a stac_version is a Catalog
        else:
            return pystac.STACObjectType.CATALOG

    return None


def identify_stac_object(json_dict: dict[str, Any]) -> STACJSONDescription:
    """Determines the STACJSONDescription of the provided JSON dict.

    Args:
        json_dict : The dict of STAC JSON to identify.

    Returns:
        STACJSONDescription: The description of the STAC object serialized in the
        given dict.
    """
    object_type = identify_stac_object_type(json_dict)

    if object_type is None:
        extra_message = f"'type' attribute is {json_dict.get('type', 'not set')}"
        raise pystac.STACTypeError(json_dict, pystac.STACObject, extra_message)

    version_range = STACVersionRange()

    stac_version = json_dict.get("stac_version")
    stac_extensions = json_dict.get("stac_extensions", None)

    if stac_version is None:
        version_range.set_min("0.8.0")
    else:
        version_range.set_to_single(stac_version)

    if stac_extensions is not None:
        version_range.set_min("0.8.0")

    if stac_extensions is None:
        stac_extensions = []

    return STACJSONDescription(object_type, version_range, stac_extensions)



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, Any, TypeAlias, TypeVar, cast

import pystac
from pystac import STACError
from pystac.link import Link
from pystac.utils import (
    HREF,
    StringEnum,
    is_absolute_href,
    make_absolute_href,
    make_posix_style,
)

if TYPE_CHECKING:
    from pystac.catalog import Catalog

S = TypeVar("S", bound="STACObject")

OptionalMediaType: TypeAlias = str | pystac.MediaType | None


class STACObjectType(StringEnum):
    CATALOG = "Catalog"
    COLLECTION = "Collection"
    ITEM = "Feature"


class STACObject(ABC):
    """A base class for other PySTAC classes that contains a variety of useful
    methods for dealing with links, copying objects, accessing extensions, and reading
    and writing files. You shouldn't use STACObject directly, but instead access this
    functionality through the implementing classes.
    """

    id: str
    """The ID of the STAC Object."""

    links: list[Link]
    """A list of :class:`~pystac.Link` objects representing all links associated with
    this STAC Object."""

    stac_extensions: list[str]
    """A list of schema URIs for STAC Extensions implemented by this STAC Object."""

    STAC_OBJECT_TYPE: STACObjectType

    _allow_parent_to_override_href: bool = True
    """Private attribute for whether parent objects should override on normalization"""

    def __init__(self, stac_extensions: list[str]) -> None:
        self.links = []
        self.stac_extensions = stac_extensions

    def validate(
        self,
        validator: pystac.validation.stac_validator.STACValidator | None = None,
    ) -> list[Any]:
        """Validate this STACObject.

        Returns a list of validation results, which depends on the validation
        implementation. For JSON Schema validation (default validator), this
        will be a list of schema URIs that were used during validation.

        Args:
            validator : A custom validator to use for validation of the object.
                If omitted, the default validator from
                :class:`~pystac.validation.RegisteredValidator`
                will be used instead.
        Raises:
            STACValidationError
        """
        import pystac.validation

        return pystac.validation.validate(self, validator=validator)

    def add_link(self, link: Link) -> None:
        """Add a link to this object's set of links.

        Args:
             link : The link to add.
        """
        link.set_owner(self)
        self.links.append(link)

    def add_links(self, links: list[Link]) -> None:
        """Add links to this object's set of links.

        Args:
             links : The links to add.
        """

        for link in links:
            self.add_link(link)

    def remove_links(self, rel: str | pystac.RelType) -> None:
        """Remove links to this object's set of links that match the given ``rel``.

        Args:
             rel : The :class:`~pystac.Link` ``rel`` to match on.
        """

        self.links = [link for link in self.links if link.rel != rel]

    def remove_hierarchical_links(self, add_canonical: bool = False) -> list[Link]:
        """Removes all hierarchical links from this object.

        See :py:const:`pystac.link.HIERARCHICAL_LINKS` for a list of all
        hierarchical links. If the object has a ``self`` href and
        ``add_canonical`` is True, a link with ``rel="canonical"`` is added.

        Args:
            add_canonical : If true, and this item has a ``self`` href, that
                href is used to build a ``canonical`` link.

        Returns:
            List[Link]: All removed links
        """
        keep = list()
        self_href = self.get_self_href()
        if add_canonical and self_href is not None:
            keep.append(
                Link("canonical", self_href, media_type=pystac.MediaType.GEOJSON)
            )
        remove = list()
        for link in self.links:
            if link.is_hierarchical():
                remove.append(link)
            else:
                keep.append(link)
        self.links = keep
        return remove

    def target_in_hierarchy(self, target: str | STACObject) -> bool:
        """Determine if target is somewhere in the hierarchical link tree of
        a STACObject.

        Args:
            target: A string or STACObject to search for

        Returns:
            bool: Returns True if the target was found in the hierarchical link tree
                for the current STACObject
        """

        def traverse(obj: str | STACObject, visited: set[str | STACObject]) -> bool:
            if obj == target:
                return True
            if isinstance(obj, str):
                return False

            new_targets = [
                link.target
                for link in obj.links
                if link.is_hierarchical() and link.target not in visited
            ]
            if target in new_targets:
                return True

            for subtree in new_targets:
                visited.add(subtree)
                if traverse(subtree, visited):
                    return True

            return False

        return traverse(self, {self})

    def get_single_link(
        self,
        rel: str | pystac.RelType | None = None,
        media_type: OptionalMediaType | Iterable[OptionalMediaType] = None,
    ) -> Link | None:
        """Get a single :class:`~pystac.Link` instance associated with this
        object.

        Args:
            rel : If set, filter links such that only those
                matching this relationship are returned.
            media_type: If set, filter the links such that only
                those matching media_type are returned. media_type can
                be a single value or a list of values.

        Returns:
            :class:`~pystac.Link` | None: First link that matches ``rel``
                and/or ``media_type``, or else the first link associated with
                this object.
        """
        if rel is None and media_type is None:
            return next(iter(self.links), None)
        if media_type and isinstance(media_type, (str, pystac.MediaType)):
            media_type = [media_type]
        return next(
            (
                link
                for link in self.links
                if (rel is None or link.rel == rel)
                and (media_type is None or link.media_type in media_type)
            ),
            None,
        )

    def get_links(
        self,
        rel: str | pystac.RelType | None = None,
        media_type: OptionalMediaType | Iterable[OptionalMediaType] = None,
    ) -> list[Link]:
        """Gets the :class:`~pystac.Link` instances associated with this object.

        Args:
            rel : If set, filter links such that only those
                matching this relationship are returned.
            media_type: If set, filter the links such that only
                those matching media_type are returned. media_type can
                be a single value or a list of values.

        Returns:
            List[:class:`~pystac.Link`]: A list of links that match ``rel`` and/
                or ``media_type`` if set, or else all links associated with this
                object.
        """
        if rel is None and media_type is None:
            return self.links
        if media_type and isinstance(media_type, (str, pystac.MediaType)):
            media_type = [media_type]
        return [
            link
            for link in self.links
            if (rel is None or link.rel == rel)
            and (media_type is None or link.media_type in media_type)
        ]

    def clear_links(self, rel: str | pystac.RelType | None = None) -> None:
        """Clears all :class:`~pystac.Link` instances associated with this object.

        Args:
            rel : If set, only clear links that match this relationship.
        """
        if rel is not None:
            self.links = [link for link in self.links if link.rel != rel]
        else:
            self.links = []

    def get_root_link(self) -> Link | None:
        """Get the :class:`~pystac.Link` representing
        the root for this object.

        Returns:
            :class:`~pystac.Link` or None: The root link for this object,
            or ``None`` if no root link is set.
        """
        return self.get_single_link(
            rel=pystac.RelType.ROOT,
            media_type=pystac.media_type.STAC_JSON,
        )

    @property
    def self_href(self) -> str:
        """Gets the absolute HREF that is represented by the ``rel == 'self'``
        :class:`~pystac.Link`.

        Raises:
            ValueError: If the self_href is not set, this method will throw
                a ValueError. Use get_self_href if there may not be an HREF set.
        """
        result = self.get_self_href()
        if result is None:
            raise ValueError(f"{self} does not have a self_href set.")
        return result

    def get_self_href(self) -> str | None:
        """Gets the absolute HREF that is represented by the ``rel == 'self'``
        :class:`~pystac.Link`.

        Returns:
            str or None: The absolute HREF of this object, or ``None`` if
            there is no self link defined.

        Note:
            A self link can exist for objects, even if the link is not read or
            written to the JSON-serialized version of the object. Any object
            read from :func:`STACObject.from_file <pystac.STACObject.from_file>` will
            have the HREF the file was read from set as it's self HREF. All self
            links have absolute (as opposed to relative) HREFs.
        """
        self_link = self.get_single_link(pystac.RelType.SELF)
        if self_link and self_link.has_target_href():
            return self_link.get_target_str()
        else:
            return None

    def set_self_href(self, href: str | None) -> None:
        """Sets the absolute HREF that is represented by the ``rel == 'self'``
        :class:`~pystac.Link`.

        Args:
            href : The absolute HREF of this object. If the given HREF
                is not absolute, it will be transformed to an absolute
                HREF based on the current working directory. If this is None
                the call will clear the self HREF link.
        """
        root_link = self.get_root_link()
        if root_link is not None and root_link.is_resolved():
            cast(pystac.Catalog, root_link.target)._resolved_objects.remove(self)

        self.remove_links(pystac.RelType.SELF)
        if href is not None:
            self.add_link(Link.self_href(href))

        if root_link is not None and root_link.is_resolved():
            cast(pystac.Catalog, root_link.target)._resolved_objects.cache(self)

    def get_root(self) -> Catalog | None:
        """Get the :class:`~pystac.Catalog` or :class:`~pystac.Collection` to
        the root for this object. The root is represented by a
        :class:`~pystac.Link` with ``rel == 'root'``.

        Returns:
            Catalog, Collection, or None:
                The root object for this object, or ``None`` if no root link is set.
        """
        root_link = self.get_root_link()
        if root_link:
            if not root_link.is_resolved():
                root_link.resolve_stac_object()
                # Use set_root, so Catalogs can merge ResolvedObjectCache instances.
                self.set_root(cast(pystac.Catalog, root_link.target))
            return cast(pystac.Catalog, root_link.target)
        else:
            return None

    def set_root(self, root: Catalog | None) -> None:
        """Sets the root :class:`~pystac.Catalog` or :class:`~pystac.Collection`
        for this object.

        Args:
            root : The root
                object to set. Passing in None will clear the root.
        """
        root_link_index = next(
            iter(
                [
                    i
                    for i, link in enumerate(self.links)
                    if link.rel == pystac.RelType.ROOT
                ]
            ),
            None,
        )

        # Remove from old root resolution cache
        if root_link_index is not None:
            root_link = self.links[root_link_index]
            if root_link.is_resolved():
                cast(pystac.Catalog, root_link.target)._resolved_objects.remove(self)

        if root is None:
            self.remove_links(pystac.RelType.ROOT)
        else:
            new_root_link = Link.root(root)
            if root_link_index is not None:
                self.links[root_link_index] = new_root_link
                new_root_link.set_owner(self)
            else:
                self.add_link(new_root_link)
            root._resolved_objects.cache(self)

    def get_parent(self) -> Catalog | None:
        """Get the :class:`~pystac.Catalog` or :class:`~pystac.Collection` to
        the parent for this object. The root is represented by a
        :class:`~pystac.Link` with ``rel == 'parent'``.

        Returns:
            Catalog, Collection, or None:
                The parent object for this object,
                or ``None`` if no root link is set.
        """
        parent_link = self.get_single_link(pystac.RelType.PARENT)
        if parent_link:
            return cast(pystac.Catalog, parent_link.resolve_stac_object().target)
        else:
            return None

    def set_parent(self, parent: Catalog | None) -> None:
        """Sets the parent :class:`~pystac.Catalog` or :class:`~pystac.Collection`
        for this object.

        Args:
            parent : The parent
                object to set. Passing in None will clear the parent.
        """

        self.remove_links(pystac.RelType.PARENT)
        if parent is not None:
            self.add_link(Link.parent(parent))

    def get_stac_objects(
        self,
        rel: str | pystac.RelType,
        typ: type[STACObject] | None = None,
        modify_links: Callable[[list[Link]], list[Link]] | None = None,
    ) -> Iterable[STACObject]:
        """Gets the :class:`STACObject` instances that are linked to
        by links with their ``rel`` property matching the passed in argument.

        Args:
            rel : The relation to match each :class:`~pystac.Link`'s
                ``rel`` property against.
            typ : If not ``None``, objects will only be yielded if they are instances of
                ``typ``.
            modify_links : A function that modifies the list of links before they are
                iterated over. For instance, this option can be used to sort the list
                so that links matching a particular pattern are earlier in the iterator.

        Returns:
            Iterable[STACObject]: A possibly empty iterable of STACObjects that are
                connected to this object through links with the given ``rel`` and are of
                type ``typ`` (if given).
        """
        links = self.links[:]
        if modify_links:
            links = modify_links(links)

        has_next_link = False
        for i in range(0, len(links)):
            link = links[i]
            if link.rel == rel:
                link.resolve_stac_object(root=self.get_root())
                if typ is None or isinstance(link.target, typ):
                    yield cast(STACObject, link.target)
            if link.rel == "next":
                has_next_link = True
        if has_next_link:
            warnings.warn(
                "This STAC object has a 'next' link, but pystac does not support "
                "pagination. If you need to paginate through responses from a STAC "
                "API, use pystac-client: https://github.com/stac-utils/pystac-client"
            )

    def save_object(
        self,
        include_self_link: bool = True,
        dest_href: str | None = None,
        stac_io: pystac.StacIO | None = None,
    ) -> None:
        """Saves this :class:`STACObject` to it's 'self' HREF.

        Args:
            include_self_link : If this is true, include the 'self' link with
                this object. Otherwise, leave out the self link.
            dest_href : Optional HREF to save the file to. If None, the object
                will be saved to the object's self href.
            stac_io: Optional instance of StacIO to use. If not provided, will use the
                instance set on the object's root if available, otherwise will use the
                default instance.


        Raises:
            STACError: If no self href is set, this error will be raised.

        Note:
            When to include a self link is described in the :stac-spec:`Use of Links
            section of the STAC best practices document
            <best-practices.md#use-of-links>`
        """
        if stac_io is None:
            root = self.get_root()
            if root is not None:
                root_stac_io = root._stac_io
                if root_stac_io is not None:
                    stac_io = root_stac_io

            if stac_io is None:
                stac_io = pystac.StacIO.default()

        if dest_href is None:
            self_href = self.get_self_href()
            if self_href is None:
                raise STACError(
                    "Self HREF must be set before saving without an explicit dest_href."
                )
            dest_href = self_href

        stac_io.save_json(dest_href, self.to_dict(include_self_link=include_self_link))

    def full_copy(
        self,
        root: Catalog | None = None,
        parent: Catalog | None = None,
    ) -> STACObject:
        """Create a full copy of this STAC object and any STAC objects linked to by
        this object.

        Args:
            root : Optional root to set as the root of the copied object,
                and any other copies that are contained by this object.
            parent : Optional parent to set as the parent of the copy
                of this object.

        Returns:
            STACObject: A full copy of this object, as well as any objects this object
                links to.
        """
        clone = self.clone()

        if root is None and isinstance(clone, pystac.Catalog):
            root = clone

        # Set the root of the STAC Object using the base class,
        # avoiding child class overrides
        # extra logic which can be incompatible with the full copy.
        STACObject.set_root(clone, cast(pystac.Catalog, root))

        if parent:
            clone.set_parent(parent)

        link_rels = set(self._object_links())
        for link in clone.links:
            if link.rel in link_rels:
                link.resolve_stac_object()
                target = cast(STACObject, link.target)
                if root is not None and target in root._resolved_objects:
                    cached_target = root._resolved_objects.get(target)
                    assert cached_target is not None
                    target = cached_target
                else:
                    target_parent = None
                    if link.rel in [
                        pystac.RelType.CHILD,
                        pystac.RelType.ITEM,
                    ] and isinstance(clone, pystac.Catalog):
                        target_parent = clone
                    copied_target = target.full_copy(root=root, parent=target_parent)
                    if root is not None:
                        root._resolved_objects.cache(copied_target)
                    target = copied_target
                if link.rel in [pystac.RelType.CHILD, pystac.RelType.ITEM]:
                    target.set_root(root)
                    if isinstance(clone, pystac.Catalog):
                        target.set_parent(clone)
                link.target = target

        return clone

    def resolve_links(self) -> None:
        """Ensure all STACObjects linked to by this STACObject are
        resolved. This is important for operations such as changing
        HREFs.

        This method mutates the entire catalog tree.
        """
        link_rels = set(self._object_links()) | {
            pystac.RelType.ROOT,
            pystac.RelType.PARENT,
        }

        for link in self.links:
            if link.rel in link_rels:
                if not link.is_resolved():
                    link.resolve_stac_object(root=self.get_root())

    @abstractmethod
    def _object_links(self) -> list[str]:
        """Inherited classes return a list of link 'rel' types that represent
        STACObjects linked to by this object (not including root, parent or self).
        This can include optional relations (which may not be present).
        """
        raise NotImplementedError

    @abstractmethod
    def to_dict(
        self, include_self_link: bool = True, transform_hrefs: bool = True
    ) -> dict[str, Any]:
        """Returns this object as a dictionary.

        Args:
            include_self_link : If True, the dict will contain a self link
                to this object. If False, the self link will be omitted.
            transform_hrefs: If True, transform the HREF of hierarchical links
                based on the type of catalog this object belongs to (if any).
                I.e. if this object belongs to a root catalog that is
                RELATIVE_PUBLISHED or SELF_CONTAINED,
                hierarchical link HREFs will be transformed to be relative to the
                catalog root.

            dict: A serialization of the object.
        """
        raise NotImplementedError

    def _repr_html_(self) -> str:
        from html import escape

        from pystac.html.jinja_env import get_jinja_env

        jinja_env = get_jinja_env()
        if jinja_env:
            template = jinja_env.get_template("JSON.jinja2")
            return str(
                template.render(
                    dict=self.to_dict(transform_hrefs=False), plain=escape(repr(self))
                )
            )
        else:
            return escape(repr(self))

    @abstractmethod
    def clone(self) -> STACObject:
        """Clones this object.

        Cloning an object will make a copy of all properties and links of the object;
        however, it will not make copies of the targets of links (i.e. it is not a
        deep copy). To copy a STACObject fully, with all linked elements also copied,
        use :func:`~pystac.STACObject.full_copy`.

        Returns:
            STACObject: The clone of this object.
        """
        raise NotImplementedError

    @classmethod
    def from_file(cls: type[S], href: HREF, stac_io: pystac.StacIO | None = None) -> S:
        """Reads a STACObject implementation from a file.

        Args:
            href : The HREF to read the object from.
            stac_io: Optional instance of StacIO to use. If not provided, will use the
                default instance.

        Returns:
            The specific STACObject implementation class that is represented
            by the JSON read from the file located at HREF.
        """
        if cls == STACObject:
            return cast(S, pystac.read_file(href))

        href = make_posix_style(href)

        if stac_io is None:
            stac_io = pystac.StacIO.default()

        if not is_absolute_href(href):
            href = make_absolute_href(href)

        d = stac_io.read_json(href)
        o = cls.from_dict(d, href=href, migrate=True, preserve_dict=False)

        # If this is a root catalog, set the root to the catalog instance.
        root_link = o.get_root_link()
        if root_link is not None:
            if not root_link.is_resolved():
                if root_link.get_absolute_href() == href:
                    o.set_root(cast(pystac.Catalog, o))
        return o

    @classmethod
    @abstractmethod
    def from_dict(
        cls: type[S],
        d: dict[str, Any],
        href: str | None = None,
        root: Catalog | None = None,
        migrate: bool = True,
        preserve_dict: bool = True,
    ) -> S:
        """Parses this STACObject from the passed in dictionary.

        Args:
            d : The dict to parse.
            href : Optional href that is the file location of the object being
                parsed.
            root : Optional root catalog for this object.
                If provided, the root of the returned STACObject will be set
                to this parameter.
            migrate: By default, STAC objects and extensions are migrated to
                their latest supported version. Set this to False to disable
                this behavior.
            preserve_dict: If False, the dict parameter ``d`` may be modified
                during this method call. Otherwise, the dict is not mutated.
                Defaults to True, which results in a deepcopy of the
                parameter. Set to False when possible to avoid the performance
                hit of a deepcopy.

        Returns:
            STACObject: The STACObject parsed from this dict.
        """
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def matches_object_type(cls, d: dict[str, Any]) -> bool:
        """Returns a boolean indicating whether the given dictionary represents a valid
        instance of this :class:`~pystac.STACObject` subclass.

        Args:
            d : A dictionary to identify
        """
        raise NotImplementedError



"""Implements the :stac-ext:`File Info Extension <file>`."""

from __future__ import annotations

from collections.abc import Iterable
from typing import Any, Generic, Literal, TypeVar, cast

from pystac import (
    Asset,
    Catalog,
    Collection,
    ExtensionTypeError,
    Item,
    Link,
    STACObjectType,
)
from pystac.extensions.base import ExtensionManagementMixin, PropertiesExtension
from pystac.extensions.hooks import ExtensionHooks
from pystac.serialization.identify import (
    OldExtensionShortIDs,
    STACJSONDescription,
    STACVersionID,
)
from pystac.utils import StringEnum, get_required, map_opt

#: Generalized version of :class:`~pystac.Asset`, :class:`~pystac.Link`,
T = TypeVar("T", Asset, Link)

SCHEMA_URI = "https://stac-extensions.github.io/file/v2.1.0/schema.json"

PREFIX = "file:"
BYTE_ORDER_PROP = PREFIX + "byte_order"
CHECKSUM_PROP = PREFIX + "checksum"
HEADER_SIZE_PROP = PREFIX + "header_size"
SIZE_PROP = PREFIX + "size"
VALUES_PROP = PREFIX + "values"
LOCAL_PATH_PROP = PREFIX + "local_path"


class ByteOrder(StringEnum):
    """List of allows values for the ``"file:byte_order"`` field defined by the
    :stac-ext:`File Info Extension <file>`."""

    LITTLE_ENDIAN = "little-endian"
    BIG_ENDIAN = "big-endian"


class MappingObject:
    """Represents a value map used by assets that are used as classification layers, and
    give details about the values in the asset and their meanings."""

    properties: dict[str, Any]

    def __init__(self, properties: dict[str, Any]) -> None:
        self.properties = properties

    def apply(self, values: list[Any], summary: str) -> None:
        """Sets the properties for this :class:`~MappingObject` instance.

        Args:
            values : The value(s) in the file. At least one array element is required.
            summary : A short description of the value(s).
        """
        self.values = values
        self.summary = summary

    @classmethod
    def create(cls, values: list[Any], summary: str) -> MappingObject:
        """Creates a new :class:`~MappingObject` instance.

        Args:
            values : The value(s) in the file. At least one array element is required.
            summary : A short description of the value(s).
        """
        m = cls({})
        m.apply(values=values, summary=summary)
        return m

    @property
    def values(self) -> list[Any]:
        """Gets or sets the list of value(s) in the file. At least one array element is
        required."""
        return get_required(self.properties.get("values"), self, "values")

    @values.setter
    def values(self, v: list[Any]) -> None:
        self.properties["values"] = v

    @property
    def summary(self) -> str:
        """Gets or sets the short description of the value(s)."""
        return cast(str, get_required(self.properties.get("summary"), self, "summary"))

    @summary.setter
    def summary(self, v: str) -> None:
        self.properties["summary"] = v

    @classmethod
    def from_dict(cls, d: dict[str, Any]) -> MappingObject:
        return cls.create(**d)

    def to_dict(self) -> dict[str, Any]:
        return self.properties


class FileExtension(
    Generic[T],
    PropertiesExtension,
    ExtensionManagementMixin[Catalog | Collection | Item],
):
    """A class that can be used to extend the properties of an :class:`~pystac.Asset`
    or :class:`~pystac.Link` with properties from the
    :stac-ext:`File Info Extension <file>`.

    To create an instance of :class:`FileExtension`, use the
    :meth:`FileExtension.ext` method. For example:

    .. code-block:: python

       >>> asset: pystac.Asset = ...
       >>> file_ext = FileExtension.ext(asset)
    """

    name: Literal["file"] = "file"

    def apply(
        self,
        byte_order: ByteOrder | None = None,
        checksum: str | None = None,
        header_size: int | None = None,
        size: int | None = None,
        values: list[MappingObject] | None = None,
        local_path: str | None = None,
    ) -> None:
        """Applies file extension properties to the extended Item.

        Args:
            byte_order : Optional byte order of integer values in the file. One of
                ``"big-endian"`` or ``"little-endian"``.
            checksum : Optional multihash for the corresponding file,
                encoded as hexadecimal (base 16) string with lowercase letters.
            header_size : Optional header size of the file, in bytes.
            size : Optional size of the file, in bytes.
            values : Optional list of :class:`~MappingObject` instances that lists the
                values that are in the file and describe their meaning. See the
                :stac-ext:`Mapping Object <file#mapping-object>` docs for an example.
                If given, at least one array element is required.
        """
        self.byte_order = byte_order
        self.checksum = checksum
        self.header_size = header_size
        self.size = size
        self.values = values
        self.local_path = local_path

    @property
    def byte_order(self) -> ByteOrder | None:
        """Gets or sets the byte order of integer values in the file. One of big-endian
        or little-endian."""
        return self._get_property(BYTE_ORDER_PROP, ByteOrder)

    @byte_order.setter
    def byte_order(self, v: ByteOrder | None) -> None:
        self._set_property(BYTE_ORDER_PROP, v)

    @property
    def checksum(self) -> str | None:
        """Get or sets the multihash for the corresponding file, encoded as hexadecimal
        (base 16) string with lowercase letters."""
        return self._get_property(CHECKSUM_PROP, str)

    @checksum.setter
    def checksum(self, v: str | None) -> None:
        self._set_property(CHECKSUM_PROP, v)

    @property
    def header_size(self) -> int | None:
        """Get or sets the header size of the file, in bytes."""
        return self._get_property(HEADER_SIZE_PROP, int)

    @header_size.setter
    def header_size(self, v: int | None) -> None:
        self._set_property(HEADER_SIZE_PROP, v)

    @property
    def local_path(self) -> str | None:
        """Get or sets a relative local path for the asset/link.

        The ``file:local_path`` field indicates a **relative** path that
        can be used by clients for different purposes to organize the
        files locally. For compatibility reasons the name-separator
        character in paths **must** be ``/`` and the Windows separator ``\\``
        is **not** allowed.
        """
        return self._get_property(LOCAL_PATH_PROP, str)

    @local_path.setter
    def local_path(self, v: str | None) -> None:
        self._set_property(LOCAL_PATH_PROP, v, pop_if_none=True)

    @property
    def size(self) -> int | None:
        """Get or sets the size of the file, in bytes."""
        return self._get_property(SIZE_PROP, int)

    @size.setter
    def size(self, v: int | None) -> None:
        self._set_property(SIZE_PROP, v)

    @property
    def values(self) -> list[MappingObject] | None:
        """Get or sets the list of :class:`~MappingObject` instances that lists the
        values that are in the file and describe their meaning. See the
        :stac-ext:`Mapping Object <file#mapping-object>` docs for an example. If given,
        at least one array element is required."""
        return map_opt(
            lambda values: [
                MappingObject.from_dict(mapping_obj) for mapping_obj in values
            ],
            self._get_property(VALUES_PROP, list[dict[str, Any]]),
        )

    @values.setter
    def values(self, v: list[MappingObject] | None) -> None:
        self._set_property(
            VALUES_PROP,
            map_opt(
                lambda values: [mapping_obj.to_dict() for mapping_obj in values], v
            ),
        )

    @classmethod
    def get_schema_uri(cls) -> str:
        return SCHEMA_URI

    @classmethod
    def ext(cls, obj: Asset | Link, add_if_missing: bool = False) -> FileExtension[T]:
        """Extends the given STAC Object with properties from the :stac-ext:`File Info
        Extension <file>`.

        This extension can be applied to instances of :class:`~pystac.Asset` or
        :class:`~pystac.Link`
        """
        if isinstance(obj, Asset):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(FileExtension[T], AssetFileExtension(obj))
        elif isinstance(obj, Link):
            cls.ensure_owner_has_extension(obj, add_if_missing)
            return cast(FileExtension[T], LinkFileExtension(obj))
        else:
            raise ExtensionTypeError(cls._ext_error_message(obj))


class AssetFileExtension(FileExtension[Asset]):
    """A concrete implementation of :class:`FileExtension` on an
    :class:`~pystac.Asset` that extends the Asset fields to include properties defined
    in the :stac-ext:`File Info Extension <file>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`FileExtension.ext` on an :class:`~pystac.Asset` to extend it.
    """

    asset_href: str
    """The ``href`` value of the :class:`~pystac.Asset` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Asset` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owner."""

    def __init__(self, asset: Asset):
        self.asset_href = asset.href
        self.properties = asset.extra_fields
        if asset.owner and hasattr(asset.owner, "properties"):
            self.additional_read_properties = [asset.owner.properties]

    def __repr__(self) -> str:
        return f"<AssetFileExtension Asset href={self.asset_href}>"


class LinkFileExtension(FileExtension[Link]):
    """A concrete implementation of :class:`FileExtension` on an
    :class:`~pystac.Link` that extends the Link fields to include properties defined
    in the :stac-ext:`File Info Extension <file>`.

    This class should generally not be instantiated directly. Instead, call
    :meth:`FileExtension.ext` on an :class:`~pystac.Link` to extend it.
    """

    link_href: str
    """The ``href`` value of the :class:`~pystac.Link` being extended."""

    properties: dict[str, Any]
    """The :class:`~pystac.Link` fields, including extension properties."""

    additional_read_properties: Iterable[dict[str, Any]] | None = None
    """If present, this will be a list containing 1 dictionary representing the
    properties of the owner."""

    def __init__(self, link: Link):
        self.link_href = link.href
        self.properties = link.extra_fields
        if link.owner and hasattr(link.owner, "properties"):
            self.additional_read_properties = [link.owner.properties]

    def __repr__(self) -> str:
        return f"<LinkFileExtension Link href={self.link_href}>"


class FileExtensionHooks(ExtensionHooks):
    schema_uri: str = SCHEMA_URI
    prev_extension_ids = {
        "file",
        "https://stac-extensions.github.io/file/v1.0.0/schema.json",
        "https://stac-extensions.github.io/file/v2.0.0/schema.json",
    }
    stac_object_types = {
        STACObjectType.ITEM,
        STACObjectType.COLLECTION,
        STACObjectType.CATALOG,
    }
    removed_fields = {
        "file:bits_per_sample",
        "file:data_type",
        "file:nodata",
        "file:unit",
    }

    def migrate(
        self, obj: dict[str, Any], version: STACVersionID, info: STACJSONDescription
    ) -> None:
        # The checksum field was previously its own extension.
        old_checksum: dict[str, str] | None = None
        if info.version_range.latest_valid_version() < "v1.0.0-rc.2":
            if OldExtensionShortIDs.CHECKSUM.value in info.extensions:
                old_item_checksum = obj["properties"].get("checksum:multihash")
                if old_item_checksum is not None:
                    if old_checksum is None:
                        old_checksum = {}
                    old_checksum["__item__"] = old_item_checksum
                for asset_key, asset in obj["assets"].items():
                    old_asset_checksum = asset.get("checksum:multihash")
                    if old_asset_checksum is not None:
                        if old_checksum is None:
                            old_checksum = {}
                        old_checksum[asset_key] = old_asset_checksum

                try:
                    obj["stac_extensions"].remove(OldExtensionShortIDs.CHECKSUM.value)
                except ValueError:
                    pass

        super().migrate(obj, version, info)

        if old_checksum is not None:
            if SCHEMA_URI not in obj["stac_extensions"]:
                obj["stac_extensions"].append(SCHEMA_URI)
            for key in old_checksum:
                if key == "__item__":
                    obj["properties"][CHECKSUM_PROP] = old_checksum[key]
                else:
                    obj["assets"][key][CHECKSUM_PROP] = old_checksum[key]

        found_fields = {}
        for asset_key, asset in obj.get("assets", {}).items():
            if values := set(asset.keys()).intersection(self.removed_fields):
                found_fields[asset_key] = values

        if found_fields:
            import warnings

            warnings.warn(
                f"Assets {list(found_fields.keys())} contain fields: "
                f"{list(set.union(*found_fields.values()))} which "
                "were removed from the file extension spec in v2.0.0. Please "
                "consult the release notes "
                "(https://github.com/stac-extensions/file/releases/tag/v2.0.0) "
                "for instructions on how to migrate these fields.",
                UserWarning,
            )


FILE_EXTENSION_HOOKS: ExtensionHooks = FileExtensionHooks()



# mypy: ignore-errors

_import_error_message = (
    "pystac-client is not installed.\n\n"
    "Please install pystac-client:\n\n"
    "  pip install pystac-client"
)

try:
    from pystac_client import *  # noqa: F403
except ImportError as e:
    if e.msg == "No module named 'pystac_client'":
        raise ImportError(_import_error_message) from e
    else:
        raise


def __getattr__(value: str):
    try:
        import pystac_client
    except ImportError as e:
        raise ImportError(_import_error_message) from e
    return getattr(pystac_client, value)



import logging

import pytest

from pystac import Asset, Catalog, Collection, Item, Link
from pystac.errors import ExtensionNotImplemented
from pystac.extensions.ext import (
    EXTENSION_NAME_MAPPING,
    EXTENSION_NAMES,
    AssetExt,
    CatalogExt,
    CollectionExt,
    ItemExt,
    LinkExt,
)
from tests.conftest import get_data_file

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger()


@pytest.fixture
def eo_ext_item() -> Item:
    ext_item_uri = get_data_file("eo/eo-landsat-example.json")
    return Item.from_file(ext_item_uri)


def test_ext_syntax_has(eo_ext_item: Item) -> None:
    assert eo_ext_item.ext.has("eo") is True
    assert eo_ext_item.ext.has("proj") is False

    assert eo_ext_item.assets["B1"].ext.has("eo") is True
    assert eo_ext_item.assets["B1"].ext.has("proj") is False


def test_ext_syntax_raises_if_ext_not_on_obj(eo_ext_item: Item) -> None:
    with pytest.raises(ExtensionNotImplemented):
        eo_ext_item.ext.proj.epsg


def test_ext_syntax_ext_can_be_added(eo_ext_item: Item) -> None:
    eo_ext_item.ext.add("proj")
    assert eo_ext_item.ext.proj.epsg is None


def test_ext_syntax_trying_to_add_invalid_ext_raises(item: Item) -> None:
    with pytest.raises(KeyError, match="Extension 'foo' is not a valid extension"):
        item.ext.add("foo")  # type: ignore


def test_ext_syntax_ext_can_be_removed(eo_ext_item: Item) -> None:
    original_n = len(eo_ext_item.stac_extensions)
    eo_ext_item.ext.remove("eo")
    with pytest.raises(
        ExtensionNotImplemented, match="Extension 'eo' is not implemented"
    ):
        eo_ext_item.ext.eo
    assert len(eo_ext_item.stac_extensions) == original_n - 1


all_link_ext_props = {a for a in dir(LinkExt) if not a.startswith("_")} - {
    "has",
    "add",
    "remove",
}
all_asset_ext_props = {a for a in dir(AssetExt) if not a.startswith("_")} - {
    "has",
    "add",
    "remove",
}
all_item_ext_props = {a for a in dir(ItemExt) if not a.startswith("_")} - {
    "has",
    "add",
    "remove",
}
all_collection_ext_props = {a for a in dir(CollectionExt) if not a.startswith("_")} - {
    "has",
    "add",
    "remove",
}
all_catalog_ext_props = {a for a in dir(CatalogExt) if not a.startswith("_")} - {
    "has",
    "add",
    "remove",
}


@pytest.mark.parametrize("name", all_link_ext_props)
def test_ext_syntax_every_prop_can_be_added_to_link(
    link: Link, name: EXTENSION_NAMES
) -> None:
    assert link.ext.has(name) is False
    link.ext.add(name)
    assert link.ext.has(name) is True
    link.ext.remove(name)
    with pytest.raises(
        ExtensionNotImplemented, match=f"Extension '{name}' is not implemented"
    ):
        getattr(link.ext, name)


@pytest.mark.parametrize("name", all_asset_ext_props)
def test_ext_syntax_every_prop_can_be_added_to_asset(
    asset: Asset, name: EXTENSION_NAMES
) -> None:
    assert asset.ext.has(name) is False
    asset.ext.add(name)
    assert asset.ext.has(name) is True
    asset.ext.remove(name)
    with pytest.raises(
        ExtensionNotImplemented, match=f"Extension '{name}' is not implemented"
    ):
        getattr(asset.ext, name)


@pytest.mark.parametrize("name", all_item_ext_props)
def test_ext_syntax_every_prop_can_be_added_to_item(
    item: Item, name: EXTENSION_NAMES
) -> None:
    assert item.ext.has(name) is False
    item.ext.add(name)
    assert item.ext.has(name) is True
    item.ext.remove(name)
    with pytest.raises(
        ExtensionNotImplemented, match=f"Extension '{name}' is not implemented"
    ):
        getattr(item.ext, name)


@pytest.mark.parametrize("name", all_collection_ext_props)
def test_ext_syntax_every_prop_can_be_added_to_collection(
    collection: Collection, name: EXTENSION_NAMES
) -> None:
    assert collection.ext.has(name) is False
    collection.ext.add(name)
    assert collection.ext.has(name) is True
    collection.ext.remove(name)
    with pytest.raises(
        ExtensionNotImplemented, match=f"Extension '{name}' is not implemented"
    ):
        getattr(collection.ext, name)


@pytest.mark.parametrize("name", all_catalog_ext_props)
def test_ext_syntax_every_prop_can_be_added_to_catalog(
    catalog: Catalog, name: EXTENSION_NAMES
) -> None:
    assert catalog.ext.has(name) is False
    catalog.ext.add(name)
    assert catalog.ext.has(name) is True
    catalog.ext.remove(name)
    with pytest.raises(
        ExtensionNotImplemented, match=f"Extension '{name}' is not implemented"
    ):
        getattr(catalog.ext, name)


def test_ext_syntax_every_name_has_a_prop() -> None:
    assert {
        *all_asset_ext_props,
        *all_item_ext_props,
        *all_collection_ext_props,
    } == set(EXTENSION_NAME_MAPPING.keys())



from typing import Any


class TemplateError(Exception):
    """Exception thrown when an error occurs during converting a template
    string into data for :class:`~pystac.layout.LayoutTemplate`
    """

    pass


class STACError(Exception):
    """A STACError is raised for errors relating to STAC, e.g. for
    invalid formats or trying to operate on a STAC that does not have
    the required information available.
    """

    pass


class STACTypeError(Exception):
    """A STACTypeError is raised when encountering a representation of
    a STAC entity that is not correct for the context; for example, if
    a Catalog JSON was read in as an Item.
    """

    def __init__(
        self,
        bad_dict: dict[str, Any],
        expected: type,
        extra_message: str | None = "",
    ):
        """
        Construct an exception with an appropriate error message from bad_dict and the
        expected that it didn't align with.

        Args:
            bad_dict: Dictionary that did not match the expected type
            expected: The expected type.
            extra_message: message that will be appended to the exception message.
        """
        message = (
            f"JSON (id = {bad_dict.get('id', 'unknown')}) does not represent"
            f" a {expected.__name__} instance."
        )
        if extra_message:
            message += " " + extra_message
        super().__init__(message)


class DuplicateObjectKeyError(Exception):
    """Raised when deserializing a JSON object containing a duplicate key."""

    pass


class ExtensionTypeError(Exception):
    """An ExtensionTypeError is raised when an extension is used against
    an object that the extension does not apply to
    """

    pass


class ExtensionAlreadyExistsError(Exception):
    """An ExtensionAlreadyExistsError is raised when extension hooks
    are registered with PySTAC if there are already hooks registered
    for an extension with the same ID."""

    pass


class ExtensionNotImplemented(Exception):
    """Attempted to extend a STAC object that does not implement the given
    extension."""


class RequiredPropertyMissing(Exception):
    """This error is raised when a required value was expected
    to be there but was missing or None. This will happen, for example,
    in an extension that has required properties, where the required
    property is missing from the extended object

    Args:
        obj: Description of the object that will have a property missing.
            Should include a __repr__ that identifies the object for the
            error message, or be a string that describes the object.
        prop: The property that is missing
    """

    def __init__(self, obj: str | Any, prop: str, msg: str | None = None) -> None:
        msg = msg or f"{repr(obj)} does not have required property {prop}"
        super().__init__(msg)


class STACLocalValidationError(Exception):
    """Schema not available locally"""


class STACValidationError(Exception):
    """Represents a validation error. Thrown by validation calls if the STAC JSON
    is invalid.

    Args:
        source : Source of the exception. Type will be determined by the
            validation implementation. For the default JsonSchemaValidator this will be
            ``jsonschema.ValidationError``.
    """

    def __init__(self, message: str, source: Any | None = None):
        super().__init__(message)
        self.source = source


class DeprecatedWarning(FutureWarning):
    """Issued when converting a dictionary to a STAC Item or Collection and the
    version extension ``deprecated`` field is present and set to ``True``."""

    pass
